# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Logging mechanisms for use in OpenHTF.

Below is an illustration of the tree of OpenHTF loggers:

+---------+
| openhtf |
+---------+
  |
  |  +--------------------------------+
  |--| Framework logs:                |
  |  | - openhtf.core.test_executor   |
  |  | - openhtf.util.threads         |
  |  | - etc. etc.                    |
  |  +--------------------------------+
  |
  |  +----------------------------------+
  +--| Test record logs:                |
     | - openhtf.test_record.<test_uid> |
     +----------------------------------+
       |
       |  +-----------------------------------------------------+
       +--| Test subsystem logs:                                |
          | - openhtf.test_record.<test_uid>.phase.<phase_name> |
          | - openhtf.test_record.<test_uid>.plug.<plug_name>   |
          +-----------------------------------------------------+

All of our logging handlers are attached to the top-level `openhtf` logger. The
other loggers in the tree have their logs propagate up to the top level.

The Test record and subsytem logs do not register with the centralized logging
hierarchy because those loggers cannot be cleaned up; instead, they use a Logger
subclass defined below that overrides the getChild function that directly
instanciates children rather than using the hierarchy.

------------------------------ Test record output ------------------------------

The test record loggers are loggers specific to a running test, with names
prefixed by `openhtf.test_record.<test_uid>`. These logs are saved only to the
record of the test with a matching test_uid. We call the other loggers in the
tree the "framework loggers". Their logs are saved to the records of all
currently running tests.

Test record logs are typically generated by test authors. Often this is done
with the logger attached to the TestApi object passed into test phases, e.g.:

  def MyLoggingPhase(test):
    test.logger.info('My log line.')

To facilitate emitting test record logs outside of test phases (without forcing
the author to pass the logger around), we provide the get_record_logger_for()
function which takes a Test UID and returns a logger, e.g.:

  from openhtf.util import logs

  class MyHelperClass(object):
    def __init__(self)
      self.test_uid = ''

    def MyRandomMethod(self):
      logs.get_record_logger_for(self.test_uid).info(
          'Log this to currently running test.')

  def MyPhase(test, helper):
    helper.MyRandomMethod()

  if __name__ == '__main__':
    helper = MyHelperClass()
    my_test = openhtf.Test(MyPhase.with_args(helper=helper))
    helper.test_uid = my_test.uid
    my_test.execute()

------------------------------ Command-line output -----------------------------

By default, logs are not sent to stdout. This is done to allow test authors to
provide a more streamlined and predictable console interface for test operators.
See the util.console_output module for tools for printing to the CLI.

During development you will probably want to run OpenHTF at a higher verbosity
level in order to view logs and full tracebacks of errors. The verbosity flag
can be used as follows:
  - Default: Logs are not printed.
  - `-v`: Logs are printed at the INFO level and up.
  - `-vv`: Logs are printed at the DEBUG level and up.

Additionally, the --quiet flag and CLI_QUIET variable from the console_output
module will override the verbosity setting and suppress all CLI output.
"""

import collections
import datetime
import logging
import os
import re
import sys
import textwrap

from openhtf.util import argv
from openhtf.util import console_output
from openhtf.util import functions
from openhtf.util import threads
import six

# The number of v's provided as command line arguments to control verbosity.
# Will be overridden if the ARG_PARSER below parses the -v argument.
CLI_LOGGING_VERBOSITY = 0

ARG_PARSER = argv.ModuleParser()
ARG_PARSER.add_argument(
    '-v', action=argv.StoreRepsInModule,
    target='%s.CLI_LOGGING_VERBOSITY' % __name__,
    help=textwrap.dedent('''\
        CLI logging verbosity. Can be repeated to increase verbosity (i.e. -v,
        -vv, -vvv).'''))

LOGGER_PREFIX = 'openhtf'
RECORD_LOGGER_PREFIX = '.'.join((LOGGER_PREFIX, 'test_record'))
RECORD_LOGGER_RE = re.compile(
    r'%s\.(?P<test_uid>[^.]*)\.?' % RECORD_LOGGER_PREFIX)
SUBSYSTEM_LOGGER_RE = re.compile(
    r'%s\.[^.]*\.(?P<subsys>plug|phase)\.(?P<id>[^.]*)' % RECORD_LOGGER_PREFIX)

_LOG_ONCE_SEEN = set()

LogRecord = collections.namedtuple(
    'LogRecord', 'level logger_name source lineno timestamp_millis message')


class HtfTestLogger(logging.Logger):
  """Custom Logger subclass that does not use the logging hierarchy.

  This subclass avoids the logging hierarchy in order to not accumulate loggers
  over the course of the test execution. The Python logging hierarchy is meant
  for module-level loggers; it does not support removing a logger from the
  hierarchy.

  Since this does not use the logging hierarchy, subloggers must be constructed
  using the parent's getChild method.
  """

  def getChild(self, suffix):
    child = HtfTestLogger('.'.join((self.name, suffix)))
    child.parent = self
    return child


def get_record_logger_for(test_uid):
  """Return the child logger associated with the specified test UID."""
  htf_logger = logging.getLogger(RECORD_LOGGER_PREFIX)
  record_logger = HtfTestLogger('.'.join(((RECORD_LOGGER_PREFIX, test_uid))))
  record_logger.parent = htf_logger
  return record_logger


def initialize_record_handler(test_uid, test_record, notify_update):
  """Initialize the record handler for a test.

  For each running test, we attach a record handler to the top-level OpenHTF
  logger. The handler will append OpenHTF logs to the test record, while
  filtering out logs that are specific to any other test run.
  """
  htf_logger = logging.getLogger(LOGGER_PREFIX)
  htf_logger.addHandler(RecordHandler(test_uid, test_record, notify_update))


def remove_record_handler(test_uid):
  handlers = logging.getLogger(LOGGER_PREFIX).handlers
  for handler in handlers:
    if isinstance(handler, RecordHandler) and handler.test_uid is test_uid:
      handlers.remove(handler)
      break


def log_once(log_func, msg, *args, **kwargs):
  """"Logs a message only once."""
  if msg not in _LOG_ONCE_SEEN:
    log_func(msg, *args, **kwargs)
    # Key on the message, ignoring args. This should fit most use cases.
    _LOG_ONCE_SEEN.add(msg)


class MacAddressLogFilter(logging.Filter):
  """A filter which redacts MAC addresses."""

  MAC_REPLACE_RE = re.compile(r"""
        ((?:[\dA-F]{2}:){3})       # 3-part prefix, f8:8f:ca means google
        (?:[\dA-F]{2}(:|\b)){3}    # the remaining octets
        """, re.IGNORECASE | re.VERBOSE)
  MAC_REPLACEMENT = r'\1<REDACTED>'

  def filter(self, record):
    if self.MAC_REPLACE_RE.search(record.getMessage()):
      # Update all the things to have no mac address in them
      if isinstance(record.msg, six.string_types):
        record.msg = self.MAC_REPLACE_RE.sub(self.MAC_REPLACEMENT, record.msg)
        record.args = tuple([
            self.MAC_REPLACE_RE.sub(self.MAC_REPLACEMENT, str(arg))
            if isinstance(arg, six.string_types)
            else arg for arg in record.args])
      else:
        record.msg = self.MAC_REPLACE_RE.sub(
            self.MAC_REPLACEMENT, record.getMessage())
    return True

# We use one shared instance of this, it has no internal state.
MAC_FILTER = MacAddressLogFilter()


class TestUidFilter(logging.Filter):
  """Exclude logs emitted by the record loggers of other tests."""

  def __init__(self, test_uid):
    super(TestUidFilter, self).__init__()
    self.test_uid = test_uid

  def filter(self, record):
    match = RECORD_LOGGER_RE.match(record.name)

    # Keep framework logs.
    if not match:
      return True

    # Exclude logs emitted by the record loggers of other tests.
    return match.group('test_uid') == self.test_uid


class KillableThreadSafeStreamHandler(logging.StreamHandler):

  def handle(self, record):
    # logging.Handler objects have an internal lock attribute that is a
    # threading.RLock instance; it can cause deadlocks in Python 2.7 when a
    # KillableThread is killed while its release method is running.
    with threads.safe_lock_release_context(self.lock):
      return super(KillableThreadSafeStreamHandler, self).handle(record)


class RecordHandler(logging.Handler):
  """A handler to save logs to an HTF TestRecord."""

  def __init__(self, test_uid, test_record, notify_update):
    super(RecordHandler, self).__init__()
    self.test_uid = test_uid
    self._test_record = test_record
    self._notify_update = notify_update
    self.addFilter(MAC_FILTER)
    self.addFilter(TestUidFilter(test_uid))

  def handle(self, record):
    # logging.Handler objects have an internal lock attribute that is a
    # threading.RLock instance; it can cause deadlocks in Python 2.7 when a
    # KillableThread is killed while its release method is running.
    with threads.safe_lock_release_context(self.lock):
      return super(RecordHandler, self).handle(record)

  def emit(self, record):
    """Save a logging.LogRecord to our test record.

    Logs carry useful metadata such as the logger name and level information.
    We capture this in a structured format in the test record to enable
    filtering by client applications.

    Args:
      record: A logging.LogRecord to record.
    """
    try:
      message = self.format(record)
      log_record = LogRecord(
          record.levelno, record.name, os.path.basename(record.pathname),
          record.lineno, int(record.created * 1000), message,
      )
      self._test_record.add_log_record(log_record)
      self._notify_update()
    except Exception:  # pylint: disable=broad-except
      self.handleError(record)


class CliFormatter(logging.Formatter):
  """Formats log messages for printing to the CLI."""

  def format(self, record):
    """Format the record as tersely as possible but preserve info."""
    super(CliFormatter, self).format(record)
    localized_time = datetime.datetime.utcfromtimestamp(record.created)
    terse_time = localized_time.strftime(u'%H:%M:%S')
    terse_level = record.levelname[0]
    terse_name = record.name.split('.')[-1]
    match = RECORD_LOGGER_RE.match(record.name)
    if match:
      # Figure out which OpenHTF subsystem the record came from.
      subsys_match = SUBSYSTEM_LOGGER_RE.match(record.name)
      if subsys_match:
        terse_name = '<{subsys}: {id}>'.format(
            subsys=subsys_match.group('subsys'),
            id=subsys_match.group('id'))
      else:
        # Fall back to using the last five characters of the test UUID.
        terse_name = '<test %s>' % match.group('test_uid')[-5:]
    return '{lvl} {time} {logger} - {msg}'.format(lvl=terse_level,
                                                  time=terse_time,
                                                  logger=terse_name,
                                                  msg=record.message)


@functions.call_once
def configure_logging():
  """One-time initialization of loggers. See module docstring for more info."""

  # Define the top-level logger.
  htf_logger = logging.getLogger(LOGGER_PREFIX)
  htf_logger.propagate = False
  htf_logger.setLevel(logging.DEBUG)

  # By default, don't print any logs to the CLI.
  if CLI_LOGGING_VERBOSITY == 0:
    htf_logger.addHandler(logging.NullHandler())
    return

  if CLI_LOGGING_VERBOSITY == 1:
    logging_level = logging.INFO
  else:
    logging_level = logging.DEBUG

  # Configure a handler to print to the CLI.
  cli_handler = KillableThreadSafeStreamHandler(stream=sys.stdout)
  cli_handler.setFormatter(CliFormatter())
  cli_handler.setLevel(logging_level)
  cli_handler.addFilter(MAC_FILTER)
  htf_logger.addHandler(cli_handler)

  # Suppress CLI logging if the --quiet flag is used, or while CLI_QUIET is set
  # in the console_output module.
  cli_handler.addFilter(console_output.CliQuietFilter())



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Thread library defining a few helpers."""

import contextlib
import ctypes
import functools
import logging
import sys
import threading
import time

import six
try:
  from six.moves import _thread
except ImportError:
  from six.moves import _dummy_thread as _thread


_LOG = logging.getLogger(__name__)


class ThreadTerminationError(SystemExit):
  """Sibling of SystemExit, but specific to thread termination."""


def safe_lock_release_context(rlock):
  if six.PY2:
    return _safe_lock_release_py2(rlock)
  # Python3 has a C-implementation of RLock, which doesn't have the thread
  # termination issues.
  return _placeholder_release_py3()

def join_timeout_increments(thread, timeout=None, increment=0.5):
  start = time.time()
  
  while thread.is_alive():
    if timeout and time.time() - start > timeout:
      break
    thread.join(increment)
  

@contextlib.contextmanager
def _placeholder_release_py3():
  yield


# pylint: disable=protected-access
@contextlib.contextmanager
def _safe_lock_release_py2(rlock):
  """Ensure that a threading.RLock is fully released for Python 2.

  The RLock release code is:
    https://github.com/python/cpython/blob/2.7/Lib/threading.py#L187

  The RLock object's release method does not release all of its state if an
  exception is raised in the middle of its operation.  There are three pieces of
  internal state that must be cleaned up:
  - owning thread ident, an integer.
  - entry count, an integer that counts how many times the current owner has
      locked the RLock.
  - internal lock, a threading.Lock instance that handles blocking.

  Args:
    rlock: threading.RLock, lock to fully release.

  Yields:
    None.
  """
  assert isinstance(rlock, threading._RLock)
  ident = _thread.get_ident()
  expected_count = 0
  if rlock._RLock__owner == ident:
    expected_count = rlock._RLock__count
  try:
    yield
  except ThreadTerminationError:
    # Check if the current thread still owns the lock by checking if we can
    # acquire the underlying lock.
    if rlock._RLock__block.acquire(0):
      # Lock is clean, so unlock and we are done.
      rlock._RLock__block.release()
    elif rlock._RLock__owner == ident and expected_count > 0:
      # The lock is still held up the stack, so make sure the count is accurate.
      if rlock._RLock__count != expected_count:
        rlock._RLock__count = expected_count
    elif rlock._RLock__owner == ident or rlock._RLock__owner is None:
      # The internal lock is still acquired, but either this thread or no thread
      # owns it, which means it needs to be hard reset.
      rlock._RLock__owner = None
      rlock._RLock__count = 0
      rlock._RLock__block.release()
    raise
# pylint: enable=protected-access


def loop(_=None, force=False):   # pylint: disable=invalid-name
  """Causes a function to loop indefinitely."""
  if not force:
    raise AttributeError(
        'threads.loop() is DEPRECATED.  If you really like this and want to '
        'keep it, file an issue at https://github.com/google/openhtf/issues '
        'and use it as @loop(force=True) for now.')

  def real_loop(fn):
    @functools.wraps(fn)
    def _proc(*args, **kwargs):
      """Wrapper to return."""
      while True:
        fn(*args, **kwargs)
    _proc.once = fn  # way for tests to invoke the function once
                     # you may need to pass in "self" since this may be unbound.
    return _proc
  return real_loop


class KillableThread(threading.Thread):
  """A thread object which handles exceptions and is able to be killed.

  Note: The reason we don't bother with arguments in thread proc is because this
  class is meant to be subclassed.  If you were to invoke this with
  target=Function then you'd lose the exception handling anyway.

  Based on recipe available at http://tomerfiliba.com/recipes/Thread2/

  Note: To fully address race conditions involved with the use of
  PyThreadState_SetAsyncExc, the GIL must be held from when the thread is
  checked to when it's async-raised. In this case, we're not doing that, and
  there remains the remote possibility that a thread identifier is reused and we
  accidentally kill the wrong thread.

  The kill method will only kill a background thread that (1) has not started
  yet or (2) is currently running its _thread_proc function.  This ensures that
  the _thread_exception and _thread_finished methods are not affected, so state
  can be properly determined.  In addition, this prevents thread termination
  during garbage collection.
  """

  def __init__(self, *args, **kwargs):
    super(KillableThread, self).__init__(*args, **kwargs)
    self._running_lock = threading.Lock()
    self._killed = threading.Event()

  def run(self):
    try:
      with self._running_lock:
        if self._killed.is_set():
          raise ThreadTerminationError()
        self._thread_proc()
    except Exception:  # pylint: disable=broad-except
      if not self._thread_exception(*sys.exc_info()):
        _LOG.critical('Thread raised an exception: %s', self.name)
        raise
    finally:
      self._thread_finished()
      _LOG.debug('Thread finished: %s', self.name)

  def _is_thread_proc_running(self):
    # Acquire the lock without blocking, though this object is fully implemented
    # in C, so we cannot specify keyword arguments.
    could_acquire = self._running_lock.acquire(0)
    if could_acquire:
      self._running_lock.release()
      return False
    return True

  def _thread_proc(self):
    """The method called when executing the thread."""

  def _thread_finished(self):
    """The method called once _thread_proc has finished."""

  def _thread_exception(self, exc_type, exc_val, exc_tb):
    """The method called if _thread_proc raises an exception.

    To suppress the exception, return True from this method.

    Args:
      exc_type: exception class.
      exc_val: exception instance of the type exc_type.
      exc_tb: traceback object for the current exception instance.

    Returns:
      True if the exception should be ignored.  The default case ignores the
      exception raised by the kill functionality.
    """
    return exc_type is ThreadTerminationError

  def kill(self):
    """Terminates the current thread by raising an error."""
    self._killed.set()
    if not self.is_alive():
      logging.debug('Cannot kill thread that is no longer running.')
      return
    if not self._is_thread_proc_running():
      logging.debug("Thread's _thread_proc function is no longer running, "
                    'will not kill; letting thread exit gracefully.')
      return
    self.async_raise(ThreadTerminationError)

  def async_raise(self, exc_type):
    """Raise the exception."""
    # Should only be called on a started thread, so raise otherwise.
    assert self.ident is not None, 'Only started threads have thread identifier'

    # If the thread has died we don't want to raise an exception so log.
    if not self.is_alive():
      _LOG.debug('Not raising %s because thread %s (%s) is not alive',
                 exc_type, self.name, self.ident)
      return

    result = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(self.ident), ctypes.py_object(exc_type))
    if result == 0 and self.is_alive():
      # Don't raise an exception an error unnecessarily if the thread is dead.
      raise ValueError('Thread ID was invalid.', self.ident)
    elif result > 1:
      # Something bad happened, call with a NULL exception to undo.
      ctypes.pythonapi.PyThreadState_SetAsyncExc(self.ident, None)
      raise RuntimeError('Error: PyThreadState_SetAsyncExc %s %s (%s) %s' % (
          exc_type, self.name, self.ident, result))


class NoneByDefaultThreadLocal(threading.local):
  """Makes thread local a bit easier to use by returning None by default.

  In general thread local sucks since you set a property on one thread and it
  isn't on another so you have to guard everything with hasattr().  This class
  just returns None if something is not set making it easier to implicitly
  check.
  """

  def __getattr__(self, _):  # pylint: disable=invalid-name
    return None


def synchronized(func):  # pylint: disable=invalid-name
  """Hold self._lock while executing func."""
  @functools.wraps(func)
  def synchronized_method(self, *args, **kwargs):
    """Wrapper to return."""
    if not hasattr(self, '_lock'):
      if func.__name__ in type(self).__dict__:
        hint = ''
      else:
        hint = (' Might be missing call to super in %s.__init__?' %
                type(self).__name__)
      raise RuntimeError('Can\'t synchronize method `%s` of %s without '
                         'attribute `_lock`.%s' %
                         (func.__name__, type(self).__name__, hint))
    with self._lock:  # pylint: disable=protected-access
      return func(self, *args, **kwargs)
  return synchronized_method



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF test logic.

Run with (your virtualenv must be activated first):

  python repeat.py

Afterwards, take a look at the output printed to stdout to get an idea of
which phases are running.

In some cases it's desirable to repeat a phase if there is a temporary error
or an undesired result.  Repeats can be triggered by returning
PhaseResult.REPEAT from a phase.  The number of repeats for a particular
phase can be limited specifying a PhaseOptions.repeat_limit.
"""

from __future__ import print_function
import openhtf
import openhtf.plugs as plugs


class FailTwicePlug(plugs.BasePlug):
  """Plug that fails twice raising an exception."""

  def __init__(self):
    self.count = 0

  def run(self):
    """Increments counter and raises an exception for first two runs."""
    self.count += 1
    print('FailTwicePlug: Run number %s' % (self.count))
    if self.count < 3:
      raise RuntimeError('Fails a couple times')

    return True


class FailAlwaysPlug(plugs.BasePlug):
  """Plug that always returns False indicating failure."""

  def __init__(self):
    self.count = 0

  def run(self):
    """Increments counter and returns False indicating failure"""
    self.count += 1
    print("FailAlwaysPlug: Run number %s" % (self.count))

    return False


# This phase demonstrates catching an exception raised in the plug and
# returning PhaseResult.REPEAT to trigger a repeat.  The phase will be run a
# total of three times: two fails followed by a success
@plugs.plug(test_plug=FailTwicePlug)
def phase_repeat(test, test_plug):
  try:
    test_plug.run()

  except:
    print("Error in phase_repeat, will retry")
    return openhtf.PhaseResult.REPEAT

  print("Completed phase_repeat")


# This phase demonstrates repeating a phase based upon a result returned from a
# plug.  In this case the plug always returns an invalid result so we re-run
# the phase.  To prevent retrying in an infinite loop, we use repeat_limit to
# limit the number of retries.
@openhtf.PhaseOptions(repeat_limit=5)
@plugs.plug(test_plug=FailAlwaysPlug)
def phase_repeat_with_limit(test, test_plug):
  result = test_plug.run()

  if not result:
    print("Invalid result in phase_repeat_with_limit, will retry")
    return openhtf.PhaseResult.REPEAT

if __name__ == '__main__':
  test = openhtf.Test(phase_repeat, phase_repeat_with_limit)
  test.execute(test_start=lambda: 'RepeatDutID')



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This module contains a class to encapsulate ADB messages.

See the following in the Android source for more details:
https://android.googlesource.com/platform/system/core/+/master/adb/protocol.txt

The ADB transport layer deals in "messages", which consist of a 24 byte header
followed (optionally) by a payload.  The header consists of 6 32 bit words which
are sent across the wire in little endian format:

struct message {
    unsigned command;       /* command identifier constant      */
    unsigned arg0;          /* first argument                   */
    unsigned arg1;          /* second argument                  */
    unsigned data_length;   /* length of payload (0 is allowed) */
    unsigned data_crc32;    /* crc32 of data payload            */
    unsigned magic;         /* command ^ 0xffffffff             */
};

Receipt of an invalid message header, corrupt message payload, or an
unrecognized command MUST result in the closing of the remote connection.  The
protocol depends on shared state and any break in the message stream will
result in state getting out of sync.

This class does not keep any of this state, but rather represents a single
message entity.  See adb_protocol.py for the stateful components.
"""

import collections
import logging
import string
import struct
import threading

from openhtf.plugs.usb import usb_exceptions
from openhtf.util import timeouts
import six

_LOG = logging.getLogger(__name__)

def make_wire_commands(*ids):
  """Assemble the commands."""
  cmd_to_wire = {
      cmd: sum(ord(c) << (i * 8) for i, c in enumerate(cmd)) for cmd in ids
  }
  wire_to_cmd = {wire: cmd for cmd, wire in six.iteritems(cmd_to_wire)}
  return cmd_to_wire, wire_to_cmd


class RawAdbMessage(collections.namedtuple('RawAdbMessage',
                                           ['cmd', 'arg0', 'arg1',
                                            'data_length', 'data_checksum',
                                            'magic'])):
  """Helper class for handling the struct -> AdbMessage mapping."""

  def to_adb_message(self, data):
    """Turn the data into an ADB message."""
    message = AdbMessage(AdbMessage.WIRE_TO_CMD.get(self.cmd),
                         self.arg0, self.arg1, data)
    if (len(data) != self.data_length or
        message.data_crc32 != self.data_checksum):
      raise usb_exceptions.AdbDataIntegrityError(
          '%s (%s) received invalid data: %s', message, self, repr(data))
    return message


class AdbTransportAdapter(object):
  """Transport adapter for reading/writing AdbMessages to another transport.

  This class handles the over-the-wire sending/receiving of AdbMessages, and
  has a utility method for ignoring messages we don't care about.  This is
  a 'transport' in that it has Read/Write methods, but those methods only
  work on AdbMessage instances, not arbitrary data, so it can't quite be
  dropped in anywhere a Transport is used.

  This class maintains its own Lock so that multiple threads can read
  AdbMessages from the same underlying transport without stealing each
  other's headers/data.
  """

  def __init__(self, transport):
    """Create an AdbTransportAdapter that writes to/reads from 'transport'."""
    self._transport = transport
    self._reader_lock = threading.Lock()
    self._writer_lock = threading.Lock()

  def __str__(self):
    trans = str(self._transport)
    return '<%s: (%s)' % (type(self).__name__, trans[1:])

  def close(self):
    """Close the connection."""
    self._transport.close()

  def write_message(self, message, timeout):
    """Send the given message over this transport.

    Args:
      message: The AdbMessage to send.
      timeout: Use this timeout for the entire write operation, it should be an
        instance of timeouts.PolledTimeout.
    """
    with self._writer_lock:
      self._transport.write(message.header, timeout.remaining_ms)

      # Use any remaining time to send the data.  Note that if we get this far,
      # we always at least try to send the data (with a minimum of 10ms timeout)
      # because we don't want the remote end to get out of sync because we sent
      # a header but no data.
      if timeout.has_expired():
        _LOG.warning('Timed out between AdbMessage header and data, sending '
                     'data anyway with 10ms timeout')
        timeout = timeouts.PolledTimeout.from_millis(10)
      self._transport.write(message.data, timeout.remaining_ms)

  def read_message(self, timeout):
    """Read an AdbMessage from this transport.

    Args:
      timeout: Timeout for the entire read operation, in the form of a
        timeouts.PolledTimeout instance.  Note that for packets with a data
        payload, two USB reads are performed.

    Returns:
      The ADB message read from the device.

    Raises:
      UsbReadFailedError: There's an error during read, including timeout.
      AdbProtocolError: A message is incorrectly formatted.
      AdbTimeoutError: timeout is already expired, or expires before we read the
        entire message, specifically between reading header and data packets.
    """
    with self._reader_lock:
      raw_header = self._transport.read(
          struct.calcsize(AdbMessage.HEADER_STRUCT_FORMAT),
          timeout.remaining_ms)
      if not raw_header:
        raise usb_exceptions.AdbProtocolError('Adb connection lost')

      try:
        raw_message = RawAdbMessage(*struct.unpack(
            AdbMessage.HEADER_STRUCT_FORMAT, raw_header))
      except struct.error as exception:
        raise usb_exceptions.AdbProtocolError(
            'Unable to unpack ADB command (%s): %s (%s)',
            AdbMessage.HEADER_STRUCT_FORMAT, raw_header, exception)

      if raw_message.data_length > 0:
        if timeout.has_expired():
          _LOG.warning('Timed out between AdbMessage header and data, reading '
                       'data anyway with 10ms timeout')
          timeout = timeouts.PolledTimeout.from_millis(10)
        data = self._transport.read(raw_message.data_length,
                                    timeout.remaining_ms)
      else:
        data = ''

      return raw_message.to_adb_message(data)

  def read_until(self, expected_commands, timeout):
    """Read AdbMessages from this transport until we get an expected command.

    The ADB protocol specifies that before a successful CNXN handshake, any
    other packets must be ignored, so this method provides the ability to
    ignore unwanted commands.  It's primarily used during the initial
    connection to the device.  See Read() for more details, including more
    exceptions that may be raised.

    Args:
      expected_commands: Iterable of expected command responses, like
          ('CNXN', 'AUTH').
      timeout: timeouts.PolledTimeout object to use for timeout.

    Returns:
      The ADB message received that matched one of expected_commands.

    Raises:
      AdbProtocolError: If timeout expires between reads, this can happen
        if we are getting spammed with unexpected commands.
    """
    msg = timeouts.loop_until_timeout_or_valid(
        timeout, lambda: self.read_message(timeout),
        lambda m: m.command in expected_commands, 0)
    if msg.command not in expected_commands:
      raise usb_exceptions.AdbTimeoutError(
          'Timed out establishing connection, waiting for: %s',
          expected_commands)
    return msg


class DebugAdbTransportAdapter(AdbTransportAdapter):
  """Debug transport adapter that logs messages read/written."""

  def __init__(self, transport):
    super(DebugAdbTransportAdapter, self).__init__(transport)
    self.messages = []
    _LOG.debug('%s logging messages', self)

  def close(self):
    _LOG.debug('%s logged messages:', self)
    for message in self.messages:
      _LOG.debug(message)
    super(DebugAdbTransportAdapter, self).close()

  def read_message(self, timeout):
    message = super(DebugAdbTransportAdapter, self).read_message(timeout)
    self.messages.append('READING: %s' % message)
    return message

  def write_message(self, message, timeout):
    self.messages.append('WRITING: %s' % message)
    super(DebugAdbTransportAdapter, self).write_message(message, timeout)


class AdbMessage(object):
  """ADB Protocol message class.

  This class encapsulates all host<->device communication for ADB.  The
  attributes of this class correspond roughly to the ADB message struct.  The
  'command' attribute of this class is a stringified version of the unsigned
  command struct value, and is one of the values in ids ('SYNC', 'CNXN', 'AUTH',
  etc).  The arg0 and arg1 attributes have different meanings depending on
  the command (see adb_protocol.py for more info).

  This class stores the 'data' associated with a message, but some messages
  have no data (data will default to '').  Additionally, reading/writing
  messages to the wire results in two reads/writes because the data is actually
  sent in a second USB transaction.  This may have implications to transport
  layers that aren't direct libusb reads/writes to a local device (ie, over
  a network).

  The 'header' attribute returns the over-the-wire format of the header for
  this message.  To send a message over the header, send its header, followed
  by its data if it has any.

  Attributes:
    header
    command
    arg0
    arg1
    data
    magic
  """

  PRINTABLE_DATA = set(string.printable) - set(string.whitespace)
  CMD_TO_WIRE, WIRE_TO_CMD = make_wire_commands('SYNC', 'CNXN', 'AUTH', 'OPEN',
                                                'OKAY', 'CLSE', 'WRTE')
  # An ADB message is 6 words in little-endian.
  HEADER_STRUCT_FORMAT = '<6I'

  def __init__(self, command, arg0=0, arg1=0, data=''):
    if command not in self.CMD_TO_WIRE:
      raise usb_exceptions.AdbProtocolError('Unrecognized ADB command: %s',
                                            command)
    self._command = self.CMD_TO_WIRE[command]
    self.arg0 = arg0
    self.arg1 = arg1
    self.data = data
    self.magic = self._command ^ 0xFFFFFFFF

  @property
  def header(self):
    """The message header."""
    return struct.pack(
        self.HEADER_STRUCT_FORMAT, self._command, self.arg0, self.arg1,
        len(self.data), self.data_crc32, self.magic)

  @property
  def command(self):
    """The ADB command."""
    return self.WIRE_TO_CMD[self._command]

  def __str__(self):
    return '<%s: %s(%s, %s): %s (%s bytes)>' % (
        type(self).__name__,
        self.command,
        self.arg0,
        self.arg1,
        ''.join(char if char in self.PRINTABLE_DATA
                else '.' for char in self.data[:64]),
        len(self.data))
  __repr__ = __str__

  @property
  def data_crc32(self):
    """Return the sum of all the data bytes.

    The "crc32" used by ADB is actually just a sum of all the bytes, but we
    name this data_crc32 to be consistent with ADB.
    """
    return sum([ord(x) for x in self.data]) & 0xFFFFFFFF



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Utils for dealing with exceptions."""

import inspect
import sys


def reraise(exc_type, message=None, *args, **kwargs):  # pylint: disable=invalid-name
  """reraises an exception for exception translation.

  This is primarily used for when you immediately reraise an exception that is
  thrown in a library, so that your client will not have to depend on various
  exceptions defined in the library implementation that is being abstracted. The
  advantage of this helper function is somewhat preserve traceback information
  although it is polluted by the reraise frame.

  Example Code:
    def A():
      raise Exception('Whoops')
    def main():
      try:
        A()
      except Exception as e:
        exceptions.reraise(ValueError)
    main()

  Traceback (most recent call last):
    File "exception.py", line 53, in <module>
      main()
    File "exception.py", line 49, in main
      reraise(ValueError)
    File "exception.py", line 47, in main
      A()
    File "exception.py", line 42, in A
      raise Exception('Whoops')
  ValueError: line 49

  When this code is run, the additional stack frames for calling A() and raising
  within A() are printed out in exception, whereas a bare exception translation
  would lose this information. As long as you ignore the reraise stack frame,
  the stack trace is okay looking.

  Generally this can be fixed by hacking on CPython to allow modification of
  traceback objects ala
  https://github.com/mitsuhiko/jinja2/blob/master/jinja2/debug.py, but this is
  fixed in Python 3 anyways and that method is the definition of hackery.

  Args:
    exc_type: (Exception) Exception class to create.
    message: (str) Optional message to place in exception instance. Usually not
      needed as the original exception probably has a message that will be
      printed out in the modified stacktrace.
    *args: Args to pass to exception constructor.
    **kwargs: Kwargs to pass to exception constructor.
  """
  last_lineno = inspect.currentframe().f_back.f_lineno
  line_msg = 'line %s: ' % last_lineno
  if message:
    line_msg += str(message)
  raise exc_type(line_msg, *args, **kwargs).raise_with_traceback(sys.exc_info()[2])



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""ADB protocol implementation.

Implements the ADB protocol as seen in Android's adb/adbd binaries, but only the
host side.  ADB communication is organized into messages (see adb_message.py).
Streams are built on top of these messages, using local and remote id numbers
to multiplex over the single channel (called the 'transport' here).

When a host connects to a device, there is first a CNXN handshake, which may
require authorization via AUTH messages.  Then, to connect to a particular
service on the device, the host sends an OPEN message containing local id by
which to refer to the resulting stream.  The classes defined here represent a
connection to a device with an AdbConnection, and an individual stream with an
AdbStream.

Multiple AdbStreams can be opened at once and used by multiple threads, and a
single AdbStream can have multiple threads reading from or writing to it, but
the latter is not recommended.

This implementation of the ADB stack looks something like:

  AdbDevice (Provides an interface similar to the adb commandline tool)
   |                 |
   V                 |
  AdbShellService    |
   |                 V
   |            AdbFilesyncService (Provides file push/pull functionality)
   |                 |
   V                 V
  AdbStream (Many may be opened simultaneously)
     |
     V
  AdbStreamTransport (Handles stream multiplexing and data queueing)
     |
     V
  AdbConnection (Only one open on a particular transport at a time)
     |
     V
  AdbMessage (Many of these are sent, comprising handshaking and AdbStreams)
     |
     V
  AdbTransportAdapter (Handles serialization/deserialization of messages)
     |
     V
  UsbHandle (Really, any transport that provides Read/Write, but usually USB)

Only AdbStream, AdbStreamTransport, and AdbConnection are implemented here.
This stack includes various utility classes that are intended for internal use
only, the classes implemented here that are intended for external users are:

  AdbStream: Intended to be used by anyone implementing their own service
    within ADB (such as shell:, tcp:, etc).  If you want to add support for
    a new service, you'll take in an AdbStream and use it for reads/writes
    from/to the remote endpoint for your service.

Example usage of a connection and stream:
  # Establish a connection to a device connected via my_usb_handle.
  connection = adb_protocol.AdbConnection.connect(my_usb_handle)

  # Read data from a 'shell:ls' destination stream and print the output.
  for line in connection.streaming_command('shell', 'ls'):
    print line
"""

import collections
import itertools
import logging
import threading

from enum import Enum

from openhtf.plugs.usb import adb_message
from openhtf.plugs.usb import usb_exceptions
from openhtf.util import argv
from openhtf.util import exceptions
from openhtf.util import timeouts
from six.moves import queue


ADB_MESSAGE_LOG = False

ARG_PARSER = argv.ModuleParser()
ARG_PARSER.add_argument('--adb_messsage_log',
                        action=argv.StoreTrueInModule,
                        target='%s.ADB_MESSAGE_LOG' % __name__,
                        help='Set to True to save all incoming and outgoing '
                        'AdbMessages and print them on Close().')

_LOG = logging.getLogger(__name__)

# Maximum amount of data in an ADB packet, we would like to raise this, but the
# remote end ignores it (ugh), we only send it to comply with the protocol.
MAX_ADB_DATA = 4096
# ADB protocol version.
ADB_VERSION = 0x01000000
# The local banner that we send to the remote device.
ADB_BANNER = 'googlex_adb'
# One greater than the maximum value for an ADB stream id.  There is a hard
# limit at 2**32, because it's an unsigned int, but we start seeing memory
# problems and the like way sooner, so lower this to catch leaking stream ids.
STREAM_ID_LIMIT = (2**16)


class AuthSigner(object):
  """Signer for use with authenticated ADB, introduced in 4.4.x/KitKat."""

  def sign(self, data):
    """Signs given data using a private key."""
    raise NotImplementedError()

  def get_public_key(self):
    """Returns the public key in PEM format without headers or newlines."""
    raise NotImplementedError()


class AdbStream(object):
  """This class represents an open ADB stream to a particular service.

  It encapsulates a local/remote id pair and any state associated with this
  particular open service connection.  This allows multiple streams to be
  open simultaneously.

  AdbStreams are thread-safe, but a write() will block until any previous
  write() has completed. As such, doing a huge write() followed by a short
  write() with a short timeout will likely result in the second write() timing
  out.

  Note that users should not instantiate this class directly, but rather should
  call open_stream() on an AdbConnection object.
  """

  def __init__(self, destination, transport):
    """Create a new ADB stream.

    Args:
      destination: String identifier for the destination of this stream.
      transport: AdbStreamTransport to use for reads/writes.
    """
    self._destination = destination
    self._transport = transport

  def is_closed(self):
    """Return True if the stream is closed."""
    return self._transport.is_closed()

  def __str__(self):
    return '<%s: (%s, %s->%s)>' % (type(self).__name__,
                                   self._destination,
                                   self._transport.local_id,
                                   self._transport.remote_id)
  __repr__ = __str__

  def write(self, data, timeout_ms=None):
    """Write data to this stream.

    Args:
      data: Data to write.
      timeout_ms: Timeout to use for the write/Ack transaction, in
        milliseconds (or as a PolledTimeout object).

    Raises:
      AdbProtocolError: If an ACK is not received.
      AdbStreamClosedError: If the stream is already closed, or gets closed
        before the write completes.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)
    # Break the data up into our transport's maxdata sized WRTE messages.
    while data:
      self._transport.write(
          data[:self._transport.adb_connection.maxdata], timeout)
      data = data[self._transport.adb_connection.maxdata:]

  def read(self, length=0, timeout_ms=None):
    """Reads data from the remote end of this stream.

    Internally, this data will have been contained in AdbMessages, but
    users of streams shouldn't need to care about the transport mechanism.

    Args:
      length: If provided, the number of bytes to read, otherwise all available
        data will be returned (at least one byte).
      timeout_ms: Time to wait for a message to come in for this stream, in
        milliseconds (or as a PolledTimeout object).

    Returns:
      Data that was read, or None if the end of the stream was reached.

    Raises:
      AdbProtocolError: Received an unexpected wonky non-stream packet (like a
        CNXN ADB message).
      AdbStreamClosedError: The stream is already closed.
      AdbTimeoutError: Timed out waiting for a message.
    """
    return self._transport.read(
        length, timeouts.PolledTimeout.from_millis(timeout_ms))

  def read_until_close(self, timeout_ms=None):
    """Yield data until this stream is closed.

    Args:
      timeout_ms: Timeout in milliseconds to keep reading (or a PolledTimeout
        object).

    Yields:
      Data read from a single call to self.read(), until the stream is closed
    or timeout is reached.

    Raises:
      AdbTimeoutError: On timeout.
    """
    while True:
      try:
        yield self.read(timeout_ms=timeout_ms)
      except usb_exceptions.AdbStreamClosedError:
        break

  def close(self, timeout_ms=100):
    """Close the stream."""
    self._transport.close(timeout_ms)


class AdbStreamTransport(object): # pylint: disable=too-many-instance-attributes
  """This class encapsulates the transport aspect of an ADB stream.

  This class handles the interface between AdbStreams and an AdbConnection,
  including the queue of messages that the AdbConnection has multiplexed and the
  stream state logic. This enables AdbStreams to use a simple Read/Write
  interface, and offloads some of the complicated state so the AdbConnection
  doesn't have to maintain it for many AdbStreams.

  Attributes:
    local_id: The local stream id for the stream using this transport.
    remote_id: The remote stream id for the stream using this transport.
    message_queue: The Queue of AdbMessages intended for this stream.
    closed: True if this transport has been closed, from either end.
  """
  ClosedState = Enum('ClosedState', ['CLOSED', 'PENDING', 'OPEN'])

  def __init__(self, adb_connection, local_id, message_queue):
    self.adb_connection = adb_connection
    self.local_id = local_id
    self.message_queue = message_queue
    # When we read the first OKAY, remote_id and closed will get updated.
    self.remote_id = None
    self.closed_state = self.ClosedState.PENDING
    # Read buffer, deque of string containing data read for this stream.
    self._read_buffer = collections.deque()
    self._buffer_size = 0  # Total bytes in _read_buffer.
    self._read_buffer_lock = threading.Lock()
    # Lock to protect against multiple simultaneous in-flight writes.
    self._write_lock = threading.Lock()
    # Some locking/events to protect against race conditions reading messages.
    self._expecting_okay = True
    self._message_received = threading.Condition()
    self._reader_lock = threading.Lock()

  def __str__(self):
    return '<%s: (%s->%s)>' % (type(self).__name__,
                               self.local_id,
                               self.remote_id)
  __repr__ = __str__

  def _set_or_check_remote_id(self, remote_id):
    """Set or check the remote id."""
    if not self.remote_id:
      assert self.closed_state == self.ClosedState.PENDING, 'Bad ClosedState!'
      self.remote_id = remote_id
      self.closed_state = self.ClosedState.OPEN
    elif self.remote_id != remote_id:
      raise usb_exceptions.AdbProtocolError(
          '%s remote-id change to %s', self, remote_id)

  def _send_command(self, command, timeout, data=''):
    """Send the given command/data over this transport.

    We do a couple sanity checks in here to be sure we can format a valid
    AdbMessage for our underlying AdbConnection, and then send it.  This method
    can be used to send any message type, and doesn't do any state tracking or
    acknowledgement checking.

    Args:
      command: The command to send, should be one of 'OKAY', 'WRTE', or 'CLSE'
      timeout: timeouts.PolledTimeout to use for this operation
      data: If provided, data to send with the AdbMessage.
    """
    if len(data) > self.adb_connection.maxdata:
      raise usb_exceptions.AdbProtocolError('Message data too long (%s>%s): %s',
                                            len(data),
                                            self.adb_connection.maxdata, data)
    if not self.remote_id:
      # If we get here, we probably missed the OKAY response to our OPEN.  We
      # should have failed earlier, but in case someone does something tricky
      # with multiple threads, we sanity check this here.
      raise usb_exceptions.AdbProtocolError('%s send before OKAY: %s',
                                            self, data)
    self.adb_connection.transport.write_message(
        adb_message.AdbMessage(command, self.local_id, self.remote_id, data),
        timeout)

  def _handle_message(self, message, handle_wrte=True):
    """Handle a message that was read for this stream.

    For each message type, this means:
      OKAY: Check id's and make sure we are expecting an OKAY.  Clear the
        self._expecting_okay flag so any pending write()'s know.
      CLSE: Set our internal state to closed.
      WRTE: Add the data read to our internal read buffer.  Note we don't
        return the actual data because it may not be this thread that needs it.

    Args:
      message: Message that was read.
      handle_wrte: If True, we can handle WRTE messages, otherwise raise.

    Raises:
      AdbProtocolError: If we get a WRTE message but handle_wrte is False.
    """
    if message.command == 'OKAY':
      self._set_or_check_remote_id(message.arg0)
      if not self._expecting_okay:
        raise usb_exceptions.AdbProtocolError(
            '%s received unexpected OKAY: %s', self, message)
      self._expecting_okay = False
    elif message.command == 'CLSE':
      self.closed_state = self.ClosedState.CLOSED
    elif not handle_wrte:
      raise usb_exceptions.AdbProtocolError(
          '%s received WRTE before OKAY/CLSE: %s', self, message)
    else:
      with self._read_buffer_lock:
        self._read_buffer.append(message.data)
        self._buffer_size += len(message.data)

  def _read_messages_until_true(self, predicate, timeout):
    """Read a message from this stream and handle it.

    This method tries to read a message from this stream, blocking until a
    message is read.  Once read, it will handle it accordingly by calling
    self._handle_message().

    This is repeated as long as predicate() returns False.  There is some
    locking used internally here so that we don't end up with multiple threads
    blocked on a call to read_for_stream when another thread has read the
    message that caused predicate() to become True.

    Args:
      predicate: Callable, keep reading messages until it returns true.  Note
        that predicate() should not block, as doing so may cause this method to
        hang beyond its timeout.
      timeout: Timeout to use for this call.

    Raises:
      AdbStreamClosedError: If this stream is already closed.
    """
    while not predicate():
      # Hold the message_received Lock while we try to acquire the reader_lock
      # and waiting on the message_received condition, to prevent another reader
      # thread from notifying the condition between us failing to acquire the
      # reader_lock and waiting on the condition.
      self._message_received.acquire()
      if self._reader_lock.acquire(False):
        try:
          # Release the message_received Lock while we do the read so other
          # threads can wait() on the condition without having to block on
          # acquiring the message_received Lock (we may have a longer timeout
          # than them, so that would be bad).
          self._message_received.release()

          # We are now the thread responsible for reading a message.  Check
          # predicate() to make sure nobody else read a message between our last
          # check and acquiring the reader Lock.
          if predicate():
            return

          # Read and handle a message, using our timeout.
          self._handle_message(
              self.adb_connection.read_for_stream(self, timeout))

          # Notify anyone interested that we handled a message, causing them to
          # check their predicate again.
          with self._message_received:
            self._message_received.notify_all()
        finally:
          self._reader_lock.release()
      else:
        # There is some other thread reading a message.  Since we are already
        # holding the message_received Lock, we can immediately do the wait.
        try:
          self._message_received.wait(timeout.remaining)
          if timeout.has_expired():
            raise usb_exceptions.AdbTimeoutError(
                '%s timed out reading messages.', self)
        finally:
          # Make sure we release this even if an exception occurred.
          self._message_received.release()

  def ensure_opened(self, timeout):
    """Ensure this stream transport was successfully opened.

    Checks to make sure we receive our initial OKAY message.  This must be
    called after creating this AdbStreamTransport and before calling read() or
    write().

    Args:
      timeout: timeouts.PolledTimeout to use for this operation.

    Returns:
      True if this stream was successfully opened, False if the service was
    not recognized by the remote endpoint.  If False is returned, then this
    AdbStreamTransport will be already closed.

    Raises:
      AdbProtocolError: If we receive a WRTE message instead of OKAY/CLSE.
    """
    self._handle_message(self.adb_connection.read_for_stream(self, timeout),
                         handle_wrte=False)
    return self.is_open()

  def is_open(self):
    """Return True if the transport layer is open."""
    return self.closed_state == self.ClosedState.OPEN

  def is_closed(self):
    """Return true if the transport layer is closed."""
    return self.closed_state == self.ClosedState.CLOSED

  def enqueue_message(self, message, timeout):
    """Add the given message to this transport's queue.

    This method also handles ACKing any WRTE messages.

    Args:
      message: The AdbMessage to enqueue.
      timeout: The timeout to use for the operation.  Specifically, WRTE
        messages cause an OKAY to be sent; timeout is used for that send.
    """
    # Ack WRTE messages immediately, handle our OPEN ack if it gets enqueued.
    if message.command == 'WRTE':
      self._send_command('OKAY', timeout=timeout)
    elif message.command == 'OKAY':
      self._set_or_check_remote_id(message.arg0)
    self.message_queue.put(message)

  def write(self, data, timeout):
    """Write data to this stream, using the given timeouts.PolledTimeout."""
    if not self.remote_id:
      raise usb_exceptions.AdbStreamClosedError(
          'Cannot write() to half-opened %s', self)
    if self.closed_state != self.ClosedState.OPEN:
      raise usb_exceptions.AdbStreamClosedError(
          'Cannot write() to closed %s', self)
    elif self._expecting_okay:
      raise usb_exceptions.AdbProtocolError(
          'Previous WRTE failed, %s in unknown state', self)

    # Make sure we only have one WRTE in flight at a time, because ADB doesn't
    # identify which WRTE it is ACK'ing when it sends the OKAY message back.
    with self._write_lock:
      self._expecting_okay = True
      self._send_command('WRTE', timeout, data)
      self._read_messages_until_true(lambda: not self._expecting_okay, timeout)

  def read(self, length, timeout):
    """Read 'length' bytes from this stream transport.

    Args:
      length: If not 0, read this many bytes from the stream, otherwise read all
        available data (at least one byte).
      timeout: timeouts.PolledTimeout to use for this read operation.

    Returns:
      The bytes read from this stream.
    """
    self._read_messages_until_true(
        lambda: self._buffer_size and self._buffer_size >= length, timeout)

    with self._read_buffer_lock:
      data, push_back = ''.join(self._read_buffer), ''
      if length:
        data, push_back = data[:length], data[length:]
      self._read_buffer.clear()
      self._buffer_size = len(push_back)
      if push_back:
        self._read_buffer.appendleft(push_back)
    return data

  def close(self, timeout_ms):
    """Close this stream, future reads/writes will fail."""
    if self.closed_state == self.ClosedState.CLOSED:
      return

    self.closed_state = self.ClosedState.CLOSED
    try:
      if not self.adb_connection.close_stream_transport(
          self, timeouts.PolledTimeout.from_millis(timeout_ms)):
        _LOG.warning('Attempt to close mystery %s', self)
    except usb_exceptions.AdbTimeoutError:
      _LOG.warning('%s close() timed out, ignoring', self)


class AdbConnection(object):
  """This class represents a connection to an ADB device.

  In the context of the ADB documentation (/system/core/adb/protocol.txt in the
  Android source), this class corresponds to a session initiated with a CONNECT
  message, likely followed by an AUTH message.

  This class does NOT represent an individual stream, identified by remote and
  local id's in the ADB documentation.  Instead, this class has an OpenStream
  method that corresponds to the OPEN message in the ADB documentation.  That
  method returns an AdbStream that can be used to write to a particular service
  on the device.

  Clients should never instantiate this class directly, but instead should use
  the connect() classmethod to open a new connection to a device.

  Attributes:
    transport: Underlying transport for this AdbConnection, usually USB.
    maxdata: Max data payload size supported by this AdbConnection.
    systemtype: System type, according to ADB's protocol.txt, one of 'device',
      'recovery', etc.
    serial: The 'serial number', as reported by ADB in the remote banner.
    banner: The 'human readable' component of the remote banner.
  """

  # pylint: disable=too-many-instance-attributes

  # AUTH constants for arg0.
  AUTH_TOKEN = 1
  AUTH_SIGNATURE = 2
  AUTH_RSAPUBLICKEY = 3

  def __init__(self, transport, maxdata, remote_banner):
    """Create an ADB connection to a device.

    Args:
      transport: AdbTransportAdapter to use for reading/writing AdbMessages
      maxdata: Max data size the remote endpoint will accept.
      remote_banner: Banner received from the remote endpoint.
    """
    try:
      self.systemtype, self.serial, self.banner = remote_banner.split(':', 2)
    except ValueError:
      raise usb_exceptions.AdbProtocolError('Received malformed banner %s',
                                            remote_banner)
    self.transport = transport
    self.maxdata = maxdata
    self._last_id_used = 0
    self._reader_lock = threading.Lock()
    self._open_lock = threading.Lock()
    # Maps local_id: AdbStreamTransport object for the relevant stream.
    self._stream_transport_map = {}
    self._stream_transport_map_lock = threading.RLock()

  def _make_stream_transport(self):
    """Create an AdbStreamTransport with a newly allocated local_id."""
    msg_queue = queue.Queue()
    with self._stream_transport_map_lock:
      # Start one past the last id we used, and grab the first available one.
      # This mimics the ADB behavior of 'increment an unsigned and let it
      # overflow', but with a check to ensure we don't reuse an id in use,
      # even though that's unlikely with 2^32 - 1 of them available.  We try
      # at most 64 id's, if we've wrapped around and there isn't one available
      # in the first 64, there's a problem, better to fail fast than hang for
      # a potentially very long time (STREAM_ID_LIMIT can be very large).
      self._last_id_used = (self._last_id_used % STREAM_ID_LIMIT) + 1
      for local_id in itertools.islice(
          itertools.chain(
              range(self._last_id_used, STREAM_ID_LIMIT),
              range(1, self._last_id_used)), 64):
        if local_id not in list(self._stream_transport_map.keys()):
          self._last_id_used = local_id
          break
      else:
        raise usb_exceptions.AdbStreamUnavailableError('Ran out of local ids!')
      # Ignore this warning - the for loop will always have at least one
      # iteration, so local_id will always be set.
      # pylint: disable=undefined-loop-variable
      stream_transport = AdbStreamTransport(self, local_id, msg_queue)
      self._stream_transport_map[local_id] = stream_transport
    return stream_transport

  def _handle_message_for_stream(self, stream_transport, message, timeout):
    """Handle an incoming message, check if it's for the given stream.

    If the message is not for the stream, then add it to the appropriate
    message queue.

    Args:
      stream_transport: AdbStreamTransport currently waiting on a message.
      message: Message to check and handle.
      timeout: Timeout to use for the operation, should be an instance of
        timeouts.PolledTimeout.

    Returns:
      The message read if it was for this stream, None otherwise.

    Raises:
      AdbProtocolError: If we receive an unexpected message type.
    """
    if message.command not in ('OKAY', 'CLSE', 'WRTE'):
      raise usb_exceptions.AdbProtocolError(
          '%s received unexpected message: %s', self, message)

    if message.arg1 == stream_transport.local_id:
      # Ack writes immediately.
      if message.command == 'WRTE':
        # Make sure we don't get a WRTE before an OKAY/CLSE message.
        if not stream_transport.remote_id:
          raise usb_exceptions.AdbProtocolError(
              '%s received WRTE before OKAY/CLSE: %s',
              stream_transport, message)
        self.transport.write_message(adb_message.AdbMessage(
            'OKAY', stream_transport.local_id, stream_transport.remote_id),
                                     timeout)
      elif message.command == 'CLSE':
        self.close_stream_transport(stream_transport, timeout)
      return message
    else:
      # Message was not for this stream, add it to the right stream's queue.
      with self._stream_transport_map_lock:
        dest_transport = self._stream_transport_map.get(message.arg1)

      if dest_transport:
        if message.command == 'CLSE':
          self.close_stream_transport(dest_transport, timeout)
        dest_transport.enqueue_message(message, timeout)
      else:
        _LOG.warning('Received message for unknown local-id: %s', message)

  def close(self):
    """Close the connection."""
    self.transport.close()

  def open_stream(self, destination, timeout_ms=None):
    """Opens a new stream to a destination service on the device.

    Not the same as the posix 'open' or any other Open methods, this
    corresponds to the OPEN message described in the ADB protocol
    documentation mentioned above.  It creates a stream (uniquely identified
    by remote/local ids) that connects to a particular service endpoint.

    Args:
      destination: The service:command string, see ADB documentation.
      timeout_ms: Timeout in milliseconds for the Open to succeed (or as a
        PolledTimeout object).

    Raises:
      AdbProtocolError: Wrong local_id sent to us, or we didn't get a ready
        response.

    Returns:
      An AdbStream object that can be used to read/write data to the specified
      service endpoint, or None if the requested service couldn't be opened.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)

    stream_transport = self._make_stream_transport()
    self.transport.write_message(
        adb_message.AdbMessage(
            command='OPEN',
            arg0=stream_transport.local_id, arg1=0,
            data=destination + '\0'),
        timeout)
    if not stream_transport.ensure_opened(timeout):
      return None
    return AdbStream(destination, stream_transport)

  def close_stream_transport(self, stream_transport, timeout):
    """Remove the given stream transport's id from our map of id's.

    If the stream id is actually removed, we send a CLSE message to let the
    remote end know (this happens when we are ack'ing a CLSE message we
    received).  The ADB protocol doesn't say this is a requirement, but ADB
    does it, so we do too.

    Args:
      stream_transport: The stream transport to close.
      timeout: Timeout on the operation.

    Returns:
      True if the id was removed and message sent, False if it was already
    missing from the stream map (already closed).
    """
    with self._stream_transport_map_lock:
      if stream_transport.local_id in self._stream_transport_map:
        del self._stream_transport_map[stream_transport.local_id]
        # If we never got a remote_id, there's no CLSE message to send.
        if stream_transport.remote_id:
          self.transport.write_message(adb_message.AdbMessage(
              'CLSE', stream_transport.local_id, stream_transport.remote_id),
                                       timeout)
        return True
    return False

  def streaming_command(self, service, command='', timeout_ms=None):
    """One complete set of packets for a single command.

    Helper function to call open_stream and yield the output.  Sends
    service:command in a new connection, reading the data for the response. All
    the data is held in memory, large responses will be slow and can fill up
    memory.

    Args:
      service: The service on the device to talk to.
      command: The command to send to the service.
      timeout_ms: Timeout for the entire command, in milliseconds (or as a
        PolledTimeout object).

    Yields:
      The data contained in the responses from the service.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)
    stream = self.open_stream('%s:%s' % (service, command), timeout)
    if not stream:
      raise usb_exceptions.AdbStreamUnavailableError(
          '%s does not support service: %s', self, service)
    for data in stream.read_until_close(timeout):
      yield data

  def read_for_stream(self, stream_transport, timeout_ms=None):
    """Attempt to read a packet for the given stream transport.

    Will read packets from self.transport until one intended for the given
    AdbStream is found.  If another thread is already reading packets, this will
    block until that thread reads a packet for this stream, or timeout expires.
    Note that this method always returns None, but if a packet was read for the
    given stream, it will have been added to that stream's message queue.

    This is somewhat tricky to do - first we check if there's a message already
    in our queue.  If not, then we try to use our AdbConnection to read a
    message for this stream.

    If some other thread is already doing reads, then read_for_stream() will sit
    in a tight loop, with a short delay, checking our message queue for a
    message from the other thread.

    Note that we must pass the queue in from the AdbStream, rather than looking
    it up in the AdbConnection's map, because the AdbConnection may have
    removed the queue from its map (while it still had messages in it).  The
    AdbStream itself maintains a reference to the queue to avoid dropping those
    messages.

    The AdbMessage read is guaranteed to be one of 'OKAY', 'WRTE', or 'CLSE'.
    If it was a WRTE message, then it will have been automatically ACK'd with an
    OKAY message, if it was a CLSE message it will have been ACK'd with a
    corresponding CLSE message, and this AdbStream will be marked as closed.

    Args:
      stream_transport: The AdbStreamTransport for the stream that is reading
        an AdbMessage from this AdbConnection.
      timeout_ms: If provided, timeout, in milliseconds, to use.  Note this
        timeout applies to this entire call, not for each individual Read, since
        there may be multiple reads if messages for other streams are read.
        This argument may be a timeouts.PolledTimeout.

    Returns:
      AdbMessage that was read, guaranteed to be one of 'OKAY', 'CLSE', or
    'WRTE' command.

    Raises:
      AdbTimeoutError: If we don't get a packet for this stream before
        timeout expires.
      AdbStreamClosedError: If the given stream has been closed.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)
    # Bail when the timeout expires, or when we no longer have the given stream
    # in our map (it may have been closed, we don't want to leave a thread
    # hanging in this loop when that happens).
    while (not timeout.has_expired() and
           stream_transport.local_id in self._stream_transport_map):
      try:
        # Block for up to 10ms to rate-limit how fast we spin.
        return stream_transport.message_queue.get(True, .01)
      except queue.Empty:
        pass

      # If someone else has the Lock, just keep checking our queue.
      if not self._reader_lock.acquire(False):
        continue

      try:
        # Now that we've acquired the Lock, we have to check the queue again,
        # just in case someone had the Lock but hadn't yet added our message
        # to the queue when we checked the first time.  Now that we have the
        # Lock ourselves, we're sure there are no potentially in-flight reads.
        try:
          return stream_transport.message_queue.get_nowait()
        except queue.Empty:
          pass

        while not timeout.has_expired():
          msg = self._handle_message_for_stream(
              stream_transport, self.transport.read_message(timeout), timeout)
          if msg:
            return msg
      finally:
        self._reader_lock.release()

    if timeout.has_expired():
      raise usb_exceptions.AdbTimeoutError(
          'Read timed out for %s', stream_transport)

    # The stream is no longer in the map, so it's closed, but check for any
    # queued messages.
    try:
      return stream_transport.message_queue.get_nowait()
    except queue.Empty:
      raise usb_exceptions.AdbStreamClosedError(
          'Attempt to read from closed or unknown %s', stream_transport)

  @classmethod
  def connect(cls, transport, rsa_keys=None, timeout_ms=1000,
              auth_timeout_ms=100):
    """Establish a new connection to a device, connected via transport.

    Args:
      transport: A transport to use for reads/writes from/to the device,
        usually an instance of UsbHandle, but really it can be anything with
        read() and write() methods.
      rsa_keys: List of AuthSigner subclass instances to be used for
        authentication. The device can either accept one of these via the sign
        method, or we will send the result of get_public_key from the first one
        if the device doesn't accept any of them.
      timeout_ms: Timeout to wait for the device to respond to our CNXN
        request.  Actual timeout may take longer if the transport object passed
        has a longer default timeout than timeout_ms, or if auth_timeout_ms is
        longer than timeout_ms and public key auth is used.  This argument may
        be a PolledTimeout object.
      auth_timeout_ms: Timeout to wait for when sending a new public key. This
        is only relevant when we send a new public key. The device shows a
        dialog and this timeout is how long to wait for that dialog. If used
        in automation, this should be low to catch such a case as a failure
        quickly; while in interactive settings it should be high to allow
        users to accept the dialog. We default to automation here, so it's low
        by default.  This argument may be a PolledTimeout object.

    Returns:
      An instance of AdbConnection that is connected to the device.

    Raises:
      usb_exceptions.DeviceAuthError: When the device expects authentication,
        but we weren't given any valid keys.
      usb_exceptions.AdbProtocolError: When the device does authentication in an
        unexpected way, or fails to respond appropriately to our CNXN request.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)
    if ADB_MESSAGE_LOG:
      adb_transport = adb_message.DebugAdbTransportAdapter(transport)
    else:
      adb_transport = adb_message.AdbTransportAdapter(transport)
    adb_transport.write_message(
        adb_message.AdbMessage(
            command='CNXN', arg0=ADB_VERSION, arg1=MAX_ADB_DATA,
            data='host::%s\0' % ADB_BANNER),
        timeout)

    msg = adb_transport.read_until(('AUTH', 'CNXN'), timeout)
    if msg.command == 'CNXN':
      return cls(adb_transport, msg.arg1, msg.data)

    # We got an AUTH response, so we have to try to authenticate.
    if not rsa_keys:
      raise usb_exceptions.DeviceAuthError(
          'Device authentication required, no keys available.')

    # Loop through our keys, signing the last 'banner' or token.
    for rsa_key in rsa_keys:
      if msg.arg0 != cls.AUTH_TOKEN:
        raise usb_exceptions.AdbProtocolError('Bad AUTH response: %s', msg)

      signed_token = rsa_key.sign(msg.data)
      adb_transport.write_message(
          adb_message.AdbMessage(
              command='AUTH', arg0=cls.AUTH_SIGNATURE, arg1=0,
              data=signed_token),
          timeout)

      msg = adb_transport.read_until(('AUTH', 'CNXN'), timeout)
      if msg.command == 'CNXN':
        return cls(adb_transport, msg.arg1, msg.data)

    # None of the keys worked, so send a public key.
    adb_transport.write_message(
        adb_message.AdbMessage(
            command='AUTH', arg0=cls.AUTH_RSAPUBLICKEY, arg1=0,
            data=rsa_keys[0].get_public_key() + '\0'),
        timeout)
    try:
      msg = adb_transport.read_until(
          ('CNXN',), timeouts.PolledTimeout.from_millis(auth_timeout_ms))
    except usb_exceptions.UsbReadFailedError as exception:
      if exception.is_timeout():
        exceptions.reraise(usb_exceptions.DeviceAuthError,
                           'Accept auth key on device, then retry.')
      raise

    # The read didn't time-out, so we got a CNXN response.
    return cls(adb_transport, msg.arg1, msg.data)



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF test demonstrating use of measurements.

Run with (your virtualenv must be activated first):

  python measurements.py

Afterwards, check out the output in measurements.json.  If you open both this
example test and that output file and compare them, you should be able to see
where measurement values end up in the output and what the corresponding code
looks like that sets them.

TODO(someone): Write these examples.
For more complex topics, see the validators.py and dimensions.py examples.

For a simpler example, see the hello_world.py example.  If the output of this
test is confusing, start with the hello_world.py output and compare it to this
test's output.

Some constraints on measurements:

  - Measurement names must be valid python variable names.  This is mostly for
    sanity, but also ensures you can access them via attribute access in phases.
    This applies *after* any with_args() substitution (not covered in this
    tutorial, see the phases.py example for more details).

  - You cannot declare the same measurement name multiple times on the same
    phase.  Technically, you *can* declare the same measurement on multiple
    phases; measurements are attached to a specific phase in the output.  This
    isn't recommended, though, because it makes it difficult to flatten a test's
    measurements, which some output formats require.
"""

# Import openhtf with an abbreviated name, as we'll be using a bunch of stuff
# from it throughout our test scripts. See __all__ at the top of
# openhtf/__init__.py for details on what's in top-of-module namespace.
import random

import openhtf as htf

# Import this output mechanism as it's the specific one we want to use.
from openhtf.output.callbacks import json_factory

# You won't normally need to import this, see validators.py example for
# more details.  It's used for the inline measurement declaration example
# below, but normally you'll only import it when you want to define custom
# measurement validators.
from openhtf.util import validators


# Simple example of measurement use, similar to hello_world.py usage.
@htf.measures(htf.Measurement('hello_world_measurement'))
def hello_phase(test):
  test.measurements.hello_world_measurement = 'Hello!'


# An alternative simpler syntax that creates the Measurement for you.
@htf.measures('hello_again_measurement')
def again_phase(test):
  test.measurements.hello_again_measurement = 'Again!'


# Multiple measurements can be specified in a single decorator, using either of
# the above syntaxes.  Technically, these syntaxes can be mixed and matched, but
# as a matter of convention you should always use one or the other within a
# single decorator call.  You'll also note that you can stack multiple
# decorations on a single phase.  This is useful if you have a handful of simple
# measurements, and then one or two with more complex declarations (see below).
@htf.measures('first_measurement', 'second_measurement')
@htf.measures(htf.Measurement('third'), htf.Measurement('fourth'))
def lots_of_measurements(test):
  test.measurements.first_measurement = 'First!'
  # Measurements can also be access via indexing rather than attributes.
  test.measurements['second_measurement'] = 'Second :('
  # This can be handy for iterating over measurements.
  for measurement in ('third', 'fourth'):
    test.measurements[measurement] = measurement + ' is the best!'


# Basic key/value measurements are handy, but we may also want to validate a
# measurement against some criteria, or specify additional information
# describing the measurement.  Validators can get quite complex, for more
# details, see the validators.py example.
@htf.measures(htf.Measurement('validated_measurement').in_range(0, 10).doc(
    'This measurement is validated.').with_units(htf.units.SECOND))
def measure_seconds(test):
  # The 'outcome' of this measurement in the test_record result will be a PASS
  # because its value passes the validator specified (0 <= 5 <= 10).
  test.measurements.validated_measurement = 5


# These additional attributes can also be specified inline as kwargs passed
# directly to the @measures decorator.  If you do so, however, you must
# specify exactly one measurement with that decorator (ie. the first argument
# must be a string containing the measurement name).  If you want to specify
# multiple measurements this way, you can stack multiple decorators.
@htf.measures('inline_kwargs', docstring='This measurement is declared inline!',
              units=htf.units.HERTZ, validators=[validators.in_range(0, 10)])
@htf.measures('another_inline', docstring='Because why not?')
def inline_phase(test):
  # This measurement will have an outcome of FAIL, because the set value of 15
  # will not pass the 0 <= x <= 10 validator.
  test.measurements.inline_kwargs = 15
  test.measurements.another_inline = 'This one is unvalidated.'

  # Let's log a message so the operator knows the test should fail.
  test.logger.info('Set inline_kwargs to a failing value, test should FAIL!')


# A multidim measurement including how to convert to a pandas dataframe and
# a numpy array.
@htf.measures(htf.Measurement('power_time_series')
              .with_dimensions('ms', 'V', 'A'))
@htf.measures(htf.Measurement('average_voltage').with_units('V'))
@htf.measures(htf.Measurement('average_current').with_units('A'))
@htf.measures(htf.Measurement('resistance').with_units('ohm').in_range(9, 11))
def multdim_measurements(test):
  # Create some fake current and voltage over time data
  for t in range(10):
    resistance = 10
    voltage = 10 + 10.0*t
    current = voltage/resistance + .01*random.random()
    dimensions = (t, voltage, current)
    test.measurements['power_time_series'][dimensions] = 0

  # When accessing your multi-dim measurement a DimensionedMeasuredValue
  # is returned.
  dim_measured_value = test.measurements['power_time_series']

  # Let's convert that to a pandas dataframe
  power_df = dim_measured_value.to_dataframe(columns=['ms', 'V', 'A', 'n/a'])
  test.logger.info('This is what a dataframe looks like:\n%s', power_df)
  test.measurements['average_voltage'] = power_df['V'].mean()

  # We can convert the dataframe to a numpy array as well
  power_array = power_df.as_matrix()
  test.logger.info('This is the same data in a numpy array:\n%s', power_array)
  test.measurements['average_current'] = power_array.mean(axis=0)[2]

  # Finally, let's estimate the resistance
  test.measurements['resistance'] = (
      test.measurements['average_voltage'] /
      test.measurements['average_current'])


if __name__ == '__main__':
  # We instantiate our OpenHTF test with the phases we want to run as args.
  test = htf.Test(hello_phase, again_phase, lots_of_measurements,
                  measure_seconds, inline_phase, multdim_measurements)

  # In order to view the result of the test, we have to output it somewhere,
  # and a local JSON file is a convenient way to do this.  Custom output
  # mechanisms can be implemented, but for now we'll just keep it simple.
  # This will always output to the same ./measurements.json file, formatted
  # slightly for human readability.
  test.add_output_callbacks(
      json_factory.OutputToJSON('./measurements.json', indent=2))

  # Unlike hello_world.py, where we prompt for a DUT ID, here we'll just
  # use an arbitrary one.
  test.execute(test_start=lambda: 'MyDutId')



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020


"""Tests in OpenHTF.

Tests are main entry point for OpenHTF tests.  In its simplest form a
test is a series of Phases that are executed by the OpenHTF framework.

"""
import argparse
import collections
import logging
import os
import sys
import textwrap
import threading
from types import LambdaType
import uuid
import weakref

import colorama
import mutablerecords

from openhtf import util
from openhtf.core import phase_descriptor
from openhtf.core import phase_executor
from openhtf.core import phase_group
from openhtf.core import test_executor
from openhtf.core import test_record

from openhtf.util import conf
from openhtf.util import console_output
from openhtf.util import logs

import six

_LOG = logging.getLogger(__name__)

conf.declare('capture_source', description=textwrap.dedent(
    '''Whether to capture the source of phases and the test module.  This
    defaults to False since this potentially reads many files and makes large
    string copies. If True, will capture docstring also.

    Set to 'true' if you want to capture your test's source.'''),
             default_value=False)

conf.declare('capture_docstring', description=textwrap.dedent(
    '''Whether to capture the docstring of phases and the test module. 
    If True, will capture docstring.

    Set to 'true' if you want to capture your test's docstring.'''),
             default_value=False)

# TODO(arsharma): Deprecate this configuration after removing the old teardown
# specification.
conf.declare('teardown_timeout_s', default_value=30, description=
             'Default timeout (in seconds) for test teardown functions; '
             'this option is deprecated and only applies to the deprecated '
             'Test level teardown function.')

def should_create_code_info():
  return conf.capture_docstring or conf.capture_source

def should_capture_source():
  return conf.capture_source

class UnrecognizedTestUidError(Exception):
  """Raised when information is requested about an unknown Test UID."""


class InvalidTestPhaseError(Exception):
  """Raised when an invalid method is decorated."""


class InvalidTestStateError(Exception):
  """Raised when an operation is attempted in an invalid state."""


def create_arg_parser(add_help=False):
  """Creates an argparse.ArgumentParser for parsing command line flags.

  If you want to add arguments, create your own with this as a parent:

  >>> parser = argparse.ArgumentParser(
          'My args title', parents=[openhtf.create_arg_parser()])
  >>> parser.parse_args()

  Args:
    add_help: boolean option passed through to arg parser.

  Returns:
    an `argparse.ArgumentParser`

  """
  parser = argparse.ArgumentParser(
      'OpenHTF-based testing',
      parents=[
          conf.ARG_PARSER,
          console_output.ARG_PARSER,
          logs.ARG_PARSER,
          phase_executor.ARG_PARSER,
      ],
      add_help=add_help)
  parser.add_argument(
      '--config-help', action='store_true',
      help='Instead of executing the test, simply print all available config '
      'keys and their description strings.')
  return parser


class Test(object):
  """An object that represents an OpenHTF test.

  Example:

    def PhaseOne(test):
      # Integrate more widgets

    def PhaseTwo(test):
      # Analyze widget integration status

    Test(PhaseOne, PhaseTwo).execute()

  Note that Test() objects *must* be created in the main thread, but can be
  .execute()'d in a separate thread.
  """

  TEST_INSTANCES = weakref.WeakValueDictionary()
  HANDLED_SIGINT_ONCE = False
  ARG_PARSER_FACTORY = create_arg_parser

  def __init__(self, *phases, _code_info_after_file=None, **metadata):
    # Some sanity checks on special metadata keys we automatically fill in.
    if 'config' in metadata:
      raise KeyError(
          'Invalid metadata key "config", it will be automatically populated.')

    self.created_time_millis = util.time_millis()
    self.last_run_time_millis = None
    self._test_options = TestOptions()
    self._lock = threading.Lock()
    self._executor = None
    self._test_desc = TestDescriptor(
        phases, test_record.CodeInfo.uncaptured(), metadata)

    if should_create_code_info():
      # First, we copy the phases with the real CodeInfo for them.
      group = self._test_desc.phase_group.load_code_info(with_source=should_capture_source())

      # Then we replace the TestDescriptor with one that stores the test
      # module's CodeInfo as well as our newly copied phases.
      if _code_info_after_file is None:
        _code_info_after_file = __file__
        
      # Extract the module level doc from the first file in the stack after _code_info_after_file
      # Unless specified, this is file calling this constructor.
      code_info = test_record.CodeInfo.for_module_from_stack(after_file=_code_info_after_file, with_source=should_capture_source())
      self._test_desc = self._test_desc._replace(
          code_info=code_info, phase_group=group)

    # Make sure configure() gets called at least once before Execute().  The
    # user might call configure() again to override options, but we don't want
    # to force them to if they want to use defaults.  For default values, see
    # the class definition of TestOptions.
    if 'test_name' in metadata:
      # Allow legacy metadata key for specifying test name.
      self.configure(name=metadata['test_name'])
    else:
      self.configure()

  @classmethod
  def from_uid(cls, test_uid):
    """Get Test by UID.

    Args:
      test_uid:  uuid for desired test.

    Returns:
      Test object for given by UID.

    Raises:
      UnrecognizedTestUidError: If the test_uid is not recognized.
    """
    test = cls.TEST_INSTANCES.get(test_uid)
    if not test:
      raise UnrecognizedTestUidError('Test UID %s not recognized' % test_uid)
    return test

  @property
  def uid(self):
    if self._executor is not None:
      return self._executor.uid

  def make_uid(self):
    """Returns the next test execution's UID.

    This identifier must be unique but trackable across invocations of
    execute(). Therefore, it's made of four parts separated by ':'
    * Process-specific (decided on process start up)
    * Test descriptor-specific (decided on descriptor creation)
    * Execution-specific (decided on test start)
    """
    return '%s:%s:%s:%s' % (os.getpid(), self.descriptor.uid,
                            uuid.uuid4().hex[:16], util.time_millis())

  @property
  def descriptor(self):
    """Static data about this test, does not change across Execute() calls."""
    return self._test_desc

  @property
  def state(self):
    """Transient state info about the currently executing test, or None."""
    with self._lock:
      if self._executor:
        return self._executor.test_state

  def get_option(self, option):
    return getattr(self._test_options, option)

  def add_output_callbacks(self, *callbacks):
    """Add the given function as an output module to this test."""
    self._test_options.output_callbacks.extend(callbacks)

  def configure(self, **kwargs):
    """Update test-wide configuration options. See TestOptions for docs."""
    # These internally ensure they are safe to call multiple times with no weird
    # side effects.
    known_args, _ = self.__class__.ARG_PARSER_FACTORY(add_help=True).parse_known_args()
    self.known_args_hook(known_args)
    logs.configure_logging()
    for key, value in six.iteritems(kwargs):
      setattr(self._test_options, key, value)

  def known_args_hook(self, known_args):
    if known_args.config_help:
      sys.stdout.write(conf.help_text)
      sys.exit(0)

  @classmethod
  def handle_sig_int(cls, *_):
    if cls.TEST_INSTANCES:
      _LOG.error('Received SIGINT, stopping all tests.')
      for test in cls.TEST_INSTANCES.values():
        test.abort_from_sig_int()
    if not cls.HANDLED_SIGINT_ONCE:
      cls.HANDLED_SIGINT_ONCE = True
      raise KeyboardInterrupt
    # Otherwise, does not raise KeyboardInterrupt to ensure that the tests are
    # cleaned up.

  def abort_from_sig_int(self):
    """Abort test execution abruptly, only in response to SIGINT."""
    with self._lock:
      _LOG.error('Aborting %s due to SIGINT', self)
      if self._executor:
        # TestState str()'s nicely to a descriptive string, so let's log that
        # just for good measure.
        _LOG.error('Test state: %s', self._executor.test_state)
        self._executor.abort()

  # TODO(arsharma): teardown_function test option is deprecated; remove this.
  def _get_running_test_descriptor(self):
    """If there is a teardown_function, wrap current descriptor with it."""
    if not self._test_options.teardown_function:
      return self._test_desc

    teardown_phase = phase_descriptor.PhaseDescriptor.wrap_or_copy(
        self._test_options.teardown_function)
    if not teardown_phase.options.timeout_s:
      teardown_phase.options.timeout_s = conf.teardown_timeout_s
    return TestDescriptor(
        phase_group.PhaseGroup(main=[self._test_desc.phase_group],
                               teardown=[teardown_phase]),
        self._test_desc.code_info, self._test_desc.metadata)

  def execute(self, test_start=None):
    """Starts the framework and executes the given test.

    Args:
      test_start: Either a trigger phase for starting the test, or a function
                  that returns a DUT ID. If neither is provided, defaults to not
                  setting the DUT ID.
    Returns:
      Boolean indicating whether the test failed (False) or passed (True).

    Raises:
      InvalidTestStateError: if this test is already being executed.
    """
    # Lock this section so we don't .stop() the executor between instantiating
    # it and .Start()'ing it, doing so does weird things to the executor state.
    with self._lock:
      # Sanity check to make sure someone isn't doing something weird like
      # trying to Execute() the same test twice in two separate threads.  We
      # hold the lock between here and Start()'ing the executor to guarantee
      # that only one thread is successfully executing the test.
      if self._executor:
        raise InvalidTestStateError('Test already running', self._executor)

      # Snapshot some things we care about and store them.
      self._test_desc.metadata['test_name'] = self._test_options.name
      self._test_desc.metadata['config'] = conf._asdict()
      self.last_run_time_millis = util.time_millis()

      if isinstance(test_start, LambdaType):
        @phase_descriptor.PhaseOptions()
        def trigger_phase(test):
          test.test_record.dut_id = test_start()
        trigger = trigger_phase
      else:
        trigger = test_start

      if should_create_code_info() and trigger:
        trigger.code_info = test_record.CodeInfo.for_function(trigger.func, with_source=should_capture_source())

      test_desc = self._get_running_test_descriptor()
      self._executor = test_executor.TestExecutor(
          test_desc, self.make_uid(), trigger, self._test_options)

      _LOG.info('Executing test: %s', self.descriptor.code_info.name)
      self.TEST_INSTANCES[self.uid] = self
      self._executor.start()

    try:
      self._executor.wait()
    except KeyboardInterrupt:
      # The SIGINT handler only raises the KeyboardInterrupt once, so only retry
      # that once.
      self._executor.wait()
      raise
    finally:
      try:
        final_state = self._executor.finalize()

        _LOG.debug('Test completed for %s, outputting now.',
                   final_state.test_record.metadata['test_name'])
        for output_cb in self._test_options.output_callbacks:
          try:
            output_cb(final_state.test_record)
          except Exception:  # pylint: disable=broad-except
            _LOG.exception(
                'Output callback %s raised; continuing anyway', output_cb)
        # Make sure the final outcome of the test is printed last and in a
        # noticeable color so it doesn't get scrolled off the screen or missed.
        if final_state.test_record.outcome == test_record.Outcome.ERROR:
          for detail in final_state.test_record.outcome_details:
            console_output.error_print(detail.description)
        else:
          colors = collections.defaultdict(lambda: colorama.Style.BRIGHT)
          colors[test_record.Outcome.PASS] = ''.join((colorama.Style.BRIGHT,
                                                      colorama.Fore.GREEN))
          colors[test_record.Outcome.FAIL] = ''.join((colorama.Style.BRIGHT,
                                                      colorama.Fore.RED))
          msg_template = 'test: {name}  outcome: {color}{outcome}{rst}'
          console_output.banner_print(msg_template.format(
              name=final_state.test_record.metadata['test_name'],
              color=colors[final_state.test_record.outcome],
              outcome=final_state.test_record.outcome.name,
              rst=colorama.Style.RESET_ALL))
      finally:
        del self.TEST_INSTANCES[self.uid]
        self._executor.close()
        self._executor = None

    return final_state.test_record.outcome == test_record.Outcome.PASS


# TODO(arsharma): Deprecate the teardown_function in favor of PhaseGroups.
class TestOptions(mutablerecords.Record('TestOptions', [], {
    'name': 'openhtf_test',
    'output_callbacks': list,
    'teardown_function': None,
    'failure_exceptions': list,
    'default_dut_id': 'UNKNOWN_DUT',
    'stop_on_first_failure': False
})):
  """Class encapsulating various tunable knobs for Tests and their defaults.

  name: The name of the test to be put into the metadata.
  output_callbacks: List of output callbacks to run, typically it's better to
      use add_output_callbacks(), but you can pass [] here to reset them.
  teardown_function: Function to run at teardown.  We pass the same arguments to
      it as a phase.
  failure_exceptions: Exceptions to cause a test FAIL instead of ERROR. When a
      test run exits early due to an exception, the run will be marked as a FAIL
      if the raised exception matches one of the types in this list. Otherwise,
      the run is marked as ERROR.
  default_dut_id: The DUT ID that will be used if the start trigger and all
      subsequent phases fail to set one.
  stop_on_first_failure: Stop Test on first failed measurement.
  """


class TestDescriptor(collections.namedtuple(
    'TestDescriptor', ['phase_group', 'code_info', 'metadata', 'uid'])):
  """An object that represents the reusable portions of an OpenHTF test.

  This object encapsulates the static test information that is set once and used
  by the framework along the way.

  Attributes:
    phase_group: The top level phase group to execute for this Test.
    metadata: Any metadata that should be associated with test records.
    code_info: Information about the module that created the Test.
    uid: UID for this test.
  """

  def __new__(cls, phases, code_info, metadata):
    group = phase_group.PhaseGroup.convert_if_not(phases)
    return super(TestDescriptor, cls).__new__(
        cls, group, code_info, metadata, uid=uuid.uuid4().hex[:16])

  @property
  def plug_types(self):
    """Returns set of plug types required by this test."""
    return {plug.cls
            for phase in self.phase_group
            for plug in phase.plugs}


class TestApi(collections.namedtuple('TestApi', [
    'logger', 'state', 'test_record', 'measurements', 'attachments',
    'attach', 'attach_from_file', 'get_measurement', 'get_attachment',
    'notify_update'])):
  """Class passed to test phases as the first argument.

  Attributes:
    dut_id: This attribute provides getter and setter access to the DUT ID
        of the device under test by the currently running openhtf.Test.  A
        non-empty DUT ID *must* be set by the end of a test, or no output
        will be produced.  It may be set via return value from a callable
        test_start argument to openhtf.Test.Execute(), or may be set in a
        test phase via this attribute.

    logger: A Python Logger instance that can be used to log to the resulting
        TestRecord.  This object supports all the usual log levels, and
        outputs to stdout (configurable) and the frontend via the Station
        API, if it's enabled, in addition to the 'log_records' attribute
        of the final TestRecord output by the running test.

    measurements: A measurements.Collection object used to get/set
        measurement values.  See util/measurements.py for more implementation
        details, but in the simple case, set measurements directly as
        attributes on this object (see examples/measurements.py for examples).

    state: A dict (initially empty) that is persisted across test phases (but
        resets for every invocation of Execute() on an openhtf.Test).  This
        can be used for any test-wide state you need to persist across phases.
        Use this with caution, however, as it is not persisted in the output
        TestRecord or displayed on the web frontend in any way.

    test_record: A reference to the output TestRecord for the currently
        running openhtf.Test.  Direct access to this attribute is *strongly*
        discouraged, but provided as a catch-all for interfaces not otherwise
        provided by TestApi.  If you find yourself using this, please file a
        feature request for an alternative at:
          https://github.com/google/openhtf/issues/new

  Callable Attributes:
    attach: Attach binary data to the test, see TestState.attach().

    attach_from_file: Attach binary data from a file, see
        TestState.attach_from_file().

    get_attachment:  Get copy of attachment contents from current or previous
        phase, see TestState.get_attachement.

    get_measurement: Get copy of a measurement from a current or previous phase,
        see TestState.get_measurement().

    notify_update: Notify any frontends of an interesting update. Typically
        this is automatically called internally when interesting things happen,
        but it can be called by the user (takes no args), for instance if
        modifying test_record directly.



  Read-only Attributes:
    attachments: Dict mapping attachment name to test_record.Attachment
        instance containing the data that was attached (and the MIME type
        that was assumed based on extension, if any).  Only attachments
        that have been attached in the current phase show up here, and this
        attribute should not be modified directly; use TestApi.attach() or
        TestApi.attach_from_file() instead.
  """

  @property
  def dut_id(self):
    return self.test_record.dut_id

  @dut_id.setter
  def dut_id(self, dut_id):
    if self.test_record.dut_id:
      self.logger.warning('Overriding previous DUT ID "%s" with "%s".',
                          self.test_record.dut_id, dut_id)
    self.test_record.dut_id = dut_id
    self.notify_update()



"""Serves a list of stations found via multicast.

Run with:
    python -m openhtf.output.servers.dashboard_server
"""

import argparse
import collections
import json
import logging
import six
import socket
import threading
import time

import sockjs.tornado
import tornado.web

from openhtf.output.servers import station_server
from openhtf.output.servers import pub_sub
from openhtf.output.servers import web_gui_server
from openhtf.output.web_gui import web_launcher
from openhtf.util import data
from openhtf.util import multicast

_LOG = logging.getLogger(__name__)

DASHBOARD_SERVER_TYPE = 'dashboard'

StationInfo = collections.namedtuple(
    'StationInfo',
    'cell host port station_id status test_description test_name')


def _discover(discover_socket, **kwargs):
  """Yields info about station servers announcing themselves via multicast."""
  query = station_server.MULTICAST_QUERY
  for host, response in discover_socket.send(query, **kwargs):
    try:
      result = json.loads(response)
    except ValueError:
      _LOG.warn('Received bad JSON over multicast from %s: %s', host, response)
    try:
      yield StationInfo(result['cell'], host, result['port'],
                        result['station_id'], 'ONLINE',
                        result.get('test_description'),
                        result['test_name'])
    except KeyError:
      if 'last_activity_time_millis' in result:
        _LOG.debug('Received old station API response on multicast. Ignoring.')
      else:
        _LOG.warn('Received bad multicast response from %s: %s', host, response)


class StationListHandler(tornado.web.RequestHandler):
  """GET endpoint for the list of available stations.

  Sends the same message provided by DashboardPubSub.
  """

  def get(self):
    self.write(DashboardPubSub.make_message())


class DashboardPubSub(pub_sub.PubSub):
  """WebSocket endpoint for the list of available stations."""
  _lock = threading.Lock()  # Required by pub_sub.PubSub.
  subscribers = set()  # Required by pub_sub.PubSub.
  last_message = None
  station_map = {}
  station_map_lock = threading.Lock()

  def on_subscribe(self, unused_info):
    """Called by the base class when a client connects."""
    if self.last_message is not None:
      self.send(self.last_message)

  @classmethod
  def update_stations(cls, station_info_list):
    """Called by the station discovery loop to update the station map."""
    with cls.station_map_lock:

      # By default, assume old stations are unreachable.
      for host_port, station_info in six.iteritems(cls.station_map):
        cls.station_map[host_port] = station_info._replace(status='UNREACHABLE')

      for station_info in station_info_list:
        host_port = '%s:%s' % (station_info.host, station_info.port)
        cls.station_map[host_port] = station_info

  @classmethod
  def publish_if_new(cls):
    """If the station map has changed, publish the new information."""
    message = cls.make_message()
    if message != cls.last_message:
      super(DashboardPubSub, cls).publish(message)
      cls.last_message = message

  @classmethod
  def make_message(cls):
    with cls.station_map_lock:
      return data.convert_to_base_types(cls.station_map)


class DashboardServer(web_gui_server.WebGuiServer):
  """Serves a list of known stations and an Angular frontend."""

  def __init__(self, port, static_files_root=None):
    dash_router = sockjs.tornado.SockJSRouter(DashboardPubSub, '/sub/dashboard')
    routes = dash_router.urls + [
        ('/station_list', StationListHandler),
    ]
    super(DashboardServer, self).__init__(routes, port, static_files_root=static_files_root)

  def _get_config(self):
    return {
        'server_type': DASHBOARD_SERVER_TYPE,
    }

  def run(self):
    _LOG.info(
        'Starting dashboard server at:\n'
        '  Local: http://localhost:{port}\n'
        '  Remote: http://{host}:{port}'
        .format(host=socket.gethostname(), port=self.port))
    super(DashboardServer, self).run()

  def stop(self):
    _LOG.info('Stopping dashboard server.')
    super(DashboardServer, self).stop()


def main(dashboard_server_class=DashboardServer):
  logging.basicConfig(level=logging.INFO)

  parser = argparse.ArgumentParser(
      description='Serves web GUI for interacting with multiple OpenHTF '
                  'stations.')
  parser.add_argument('--discovery-interval-s', type=int, default=1,
                      help='Seconds between station discovery attempts.')
  parser.add_argument('--launch-web-gui', default=True, action="store_true",
                      help='Whether to automatically open web GUI.')
  parser.add_argument('--no-launch-web-gui', dest="launch_web_gui",
                      action="store_false",
                      help='Whether to automatically open web GUI.')
  parser.add_argument('--dashboard-server-port', type=int, default=12000,
                      help='Port on which to serve the dashboard server.')

  # These have default values in openhtf.util.multicast.py.
  parser.add_argument('--station-discovery-address', type=str)
  parser.add_argument('--station-discovery-port', type=int)
  parser.add_argument('--station-discovery-ttl', type=int)
  parser.add_argument(
      '--no-local-only',
      action='store_false',
      default=True,
      dest='station_discovery_local_only',
      help=('Whether to discover only local stations.'))

  args = parser.parse_args()

  with dashboard_server_class(args.dashboard_server_port) as server:

    if args.launch_web_gui:
      url = 'http://localhost:%s' % (server.port,)
      try:
        web_launcher.launch(url)
      except Exception:  # pylint: disable=broad-except
        _LOG.exception('Problem launching web gui')

    # Make kwargs from command line arguments.
    multicast_kwargs = {
        attr: getattr(args, 'station_discovery_%s' % attr)
        for attr in ('address', 'port', 'ttl', 'local_only')
        if getattr(args, 'station_discovery_%s' % attr) is not None
    }

    _LOG.info('Starting station discovery.')

    # Exit on CTRL+C.
    discover_socket = multicast.MulticastSendSocket()
    while True:
      stations = _discover(discover_socket, **multicast_kwargs)
      DashboardPubSub.update_stations(list(stations))
      DashboardPubSub.publish_if_new()
      time.sleep(args.discovery_interval_s)


if __name__ == '__main__':
  main()



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example plugs for OpenHTF."""

import openhtf.plugs as plugs
from openhtf.util import conf


conf.declare('example_plug_increment_size', default_value=1,
             description='increment constant for example plug.')


class ExamplePlug(plugs.BasePlug):   # pylint: disable=no-init
  """Example of a simple plug.

  This plug simply keeps a value and increments it each time increment() is
  called.  You'll notice a few paradigms here:

    - @conf.inject_positional_args
      This is generally a good way to pass in any configuration that your
      plug needs, such as an IP address or serial port to connect to.  If
      You want to use your plug outside of the OpenHTF framework, you can
      still manually instantiate it, but you must pass the arguments by
      keyword (as a side effect of the way inject_positional_args is
      implemented).

      For example, if you had no openhtf.conf loaded, you could do this:
        my_plug = ExamplePlug(example_plug_increment_size=4)

    - tearDown()
      This method will be called automatically by the OpenHTF framework at
      the end of test execution.  Here is a good place to do any close()
      calls or similar resource cleanup that you need to do.  In this case,
      we don't have anything to do, so we simply log a message so you can
      see when it gets called.

    - Persistent 'value'
      You'll notice that value is an instance attribute, not a class
      attribute.  This is because plugs are instantiated once at the
      beginning of the test, and then the same instance is passed into
      all test phases that use that plug type.  Because of this, you
      don't have to do anything special to maintain state within a plug
      across phases.

      This does imply, however, that if you *want* per-phase tearDown()
      semantics, you have to implement them manually.  The recommended
      way to do this is to make your plug support Python's context
      manager interface (__enter__ and __exit__), and then access it via
      a with: block at the beginning of every phase where it is used.
  """

  @conf.inject_positional_args
  def __init__(self, example_plug_increment_size):
    self.increment_size = example_plug_increment_size
    self.value = 0

  def __str__(self):
    return '<%s: %s>' % (type(self).__name__, self.value)

  def tearDown(self):
    """Tear down the plug instance."""
    self.logger.info('Tearing down %s', self)

  def increment(self):
    """Increment our value, return the previous value."""
    self.value += self.increment_size
    return self.value - self.increment_size


class ExampleFrontendAwarePlug(plugs.FrontendAwareBasePlug):
  """Example of a simple frontend-aware plug.

  A frontend-aware plug is a plug that agrees to call self.notify_update()
  anytime its state changes. The state should be returned by self._asdict().
  This allows frontends such as openhtf.output.web_gui to receive updates to the
  plug's state in real time.

  See also:
    - openhtf.plugs.FrontendAwareBasePlug
    - openhtf.plugs.user_input.UserInput
  """
  def __init__(self):
    super(ExampleFrontendAwarePlug, self).__init__()
    self.value = 0

  def _asdict(self):
    return {'value': self.value}

  def increment(self):
    """Increment our value."""
    self.value += 1
    self.notify_update()



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF test demonstrating use of checkpoints and measurements."""

import time

import openhtf as htf
from openhtf.output.callbacks import console_summary
from openhtf.output.callbacks import json_factory
from openhtf.util import checkpoints

from openhtf.examples import measurements as measurements_example


@htf.measures(htf.Measurement('fixed_time').in_range(0, 10).doc(
    'This is going to fail validation.').with_units(htf.units.SECOND))
def failing_phase(test):
  # The 'outcome' of this measurement in the test_record result will be a FAIL
  # because its value fails the validator specified (0 <= 5 <= 10).
  test.measurements.fixed_time = 20


def long_running_phase(test):
  # A long running phase could be something like a hardware burn-in.  This
  # phase should not run if previous phases have failed, so we make sure
  # checkpoint phase is run right before this phase.
  for i in range(10):
    test.logger.info('Still running....')
    time.sleep(10)
  test.logger.info('Done with long_running_phase')

if __name__ == '__main__':
  # We instantiate our OpenHTF test with the phases we want to run as args.
  test = htf.Test(
      measurements_example.hello_phase,
      measurements_example.again_phase,
      failing_phase,
      measurements_example.lots_of_measurements,
      checkpoints.checkpoint(),
      long_running_phase,
  )

  # In order to view the result of the test, we have to output it somewhere,
  # outputting to console is an easy way to do this.
  test.add_output_callbacks(
      console_summary.ConsoleSummary()
  )

  # The complete summary is viable in json, including the measurements
  # included in measurements_example.lots_of_measurements.
  test.add_output_callbacks(
      json_factory.OutputToJSON('./checkpoints.json', indent=2))

  # Unlike hello_world.py, where we prompt for a DUT ID, here we'll just
  # use an arbitrary one.
  test.execute(test_start=lambda: 'MyDutId')



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openhtf/output/proto/mfg_event.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from openhtf.output.proto import assembly_event_pb2 as openhtf_dot_output_dot_proto_dot_assembly__event__pb2
from openhtf.output.proto import test_runs_pb2 as openhtf_dot_output_dot_proto_dot_test__runs__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='openhtf/output/proto/mfg_event.proto',
  package='openhtf',
  syntax='proto2',
  serialized_pb=_b('\n$openhtf/output/proto/mfg_event.proto\x12\x07openhtf\x1a)openhtf/output/proto/assembly_event.proto\x1a$openhtf/output/proto/test_runs.proto\"\xb4\x05\n\x08MfgEvent\x12\x14\n\ndut_serial\x18\x01 \x01(\tH\x00\x12&\n\x03lot\x18\x02 \x01(\x0b\x32\x17.openhtf.MfgEvent.ByLotH\x00\x12\x15\n\rstart_time_ms\x18\x03 \x02(\x03\x12\x13\n\x0b\x65nd_time_ms\x18\x04 \x01(\x03\x12\x13\n\x0btester_name\x18\x05 \x02(\t\x12\x11\n\ttest_name\x18\x06 \x02(\t\x12\x14\n\x0ctest_version\x18\x14 \x01(\t\x12\x18\n\x10test_description\x18\x15 \x01(\t\x12\x15\n\rtest_run_name\x18\x16 \x01(\t\x12-\n\x0btest_status\x18\x08 \x02(\x0e\x32\x0f.openhtf.Status:\x07\x43REATED\x12/\n\x0f\x61ssembly_events\x18\n \x03(\x0b\x32\x16.openhtf.AssemblyEvent\x12 \n\x07timings\x18\x0b \x03(\x0b\x32\x0f.openhtf.Timing\x12\x1e\n\x06phases\x18\x0c \x03(\x0b\x32\x0e.openhtf.Phase\x12\x17\n\x0f\x66ramework_build\x18\r \x01(\t\x12)\n\x0bmeasurement\x18\x0e \x03(\x0b\x32\x14.openhtf.Measurement\x12,\n\nattachment\x18\x0f \x03(\x0b\x32\x18.openhtf.EventAttachment\x12-\n\ttest_logs\x18\x10 \x03(\x0b\x32\x1a.openhtf.TestRunLogMessage\x12+\n\rfailure_codes\x18\x11 \x03(\x0b\x32\x14.openhtf.FailureCode\x12\x15\n\roperator_name\x18\x12 \x01(\t\x12\x11\n\tpart_tags\x18\x13 \x03(\t\x1a.\n\x05\x42yLot\x12\x12\n\nlot_number\x18\x01 \x02(\t\x12\x11\n\tlot_index\x18\x02 \x01(\tB\x05\n\x03\x64ut\"\xa6\x03\n\x0bMeasurement\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x1f\n\x06status\x18\x02 \x01(\x0e\x32\x0f.openhtf.Status\x12\x13\n\x0b\x64\x65scription\x18\x06 \x01(\t\x12\x11\n\timportant\x18\x12 \x01(\x08\x12\x15\n\rparameter_tag\x18\x10 \x03(\t\x12\x15\n\rnumeric_value\x18\x0b \x01(\x01\x12\x17\n\x0fnumeric_minimum\x18\x0c \x01(\x01\x12\x17\n\x0fnumeric_maximum\x18\r \x01(\x01\x12\x12\n\ntext_value\x18\x0e \x01(\t\x12\x15\n\rexpected_text\x18\x0f \x01(\t\x12\x13\n\x0bis_optional\x18\x11 \x01(\x08\x12\x17\n\x0fset_time_millis\x18\x13 \x01(\x03\x12*\n\tunit_code\x18\x14 \x01(\x0e\x32\x17.openhtf.Units.UnitCode\x12\x18\n\x10\x63ustom_unit_code\x18\x16 \x01(\t\x12\x1a\n\x12\x63ustom_unit_suffix\x18\x17 \x01(\t\x12\x1d\n\x15\x61ssociated_attachment\x18\x15 \x01(\t*\x06\x08\x88\'\x10\xd0(\"\xc4\x01\n\x0f\x45ventAttachment\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x14\n\x0cvalue_binary\x18\x02 \x01(\x0c\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x17\n\x0fset_time_millis\x18\x04 \x01(\x03\x12\x15\n\rparameter_tag\x18\x05 \x03(\t\x12\'\n\x04type\x18\x06 \x01(\x0e\x32\x17.openhtf.InformationTagH\x00\x12\x13\n\tmime_type\x18\x07 \x01(\tH\x00\x42\n\n\x08\x64\x61tatype')
  ,
  dependencies=[openhtf_dot_output_dot_proto_dot_assembly__event__pb2.DESCRIPTOR,openhtf_dot_output_dot_proto_dot_test__runs__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_MFGEVENT_BYLOT = _descriptor.Descriptor(
  name='ByLot',
  full_name='openhtf.MfgEvent.ByLot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lot_number', full_name='openhtf.MfgEvent.ByLot.lot_number', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lot_index', full_name='openhtf.MfgEvent.ByLot.lot_index', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=770,
  serialized_end=816,
)

_MFGEVENT = _descriptor.Descriptor(
  name='MfgEvent',
  full_name='openhtf.MfgEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dut_serial', full_name='openhtf.MfgEvent.dut_serial', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lot', full_name='openhtf.MfgEvent.lot', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='start_time_ms', full_name='openhtf.MfgEvent.start_time_ms', index=2,
      number=3, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='end_time_ms', full_name='openhtf.MfgEvent.end_time_ms', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tester_name', full_name='openhtf.MfgEvent.tester_name', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_name', full_name='openhtf.MfgEvent.test_name', index=5,
      number=6, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_version', full_name='openhtf.MfgEvent.test_version', index=6,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_description', full_name='openhtf.MfgEvent.test_description', index=7,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_run_name', full_name='openhtf.MfgEvent.test_run_name', index=8,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_status', full_name='openhtf.MfgEvent.test_status', index=9,
      number=8, type=14, cpp_type=8, label=2,
      has_default_value=True, default_value=5,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assembly_events', full_name='openhtf.MfgEvent.assembly_events', index=10,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timings', full_name='openhtf.MfgEvent.timings', index=11,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='phases', full_name='openhtf.MfgEvent.phases', index=12,
      number=12, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='framework_build', full_name='openhtf.MfgEvent.framework_build', index=13,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='measurement', full_name='openhtf.MfgEvent.measurement', index=14,
      number=14, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attachment', full_name='openhtf.MfgEvent.attachment', index=15,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_logs', full_name='openhtf.MfgEvent.test_logs', index=16,
      number=16, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='failure_codes', full_name='openhtf.MfgEvent.failure_codes', index=17,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='operator_name', full_name='openhtf.MfgEvent.operator_name', index=18,
      number=18, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='part_tags', full_name='openhtf.MfgEvent.part_tags', index=19,
      number=19, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_MFGEVENT_BYLOT, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='dut', full_name='openhtf.MfgEvent.dut',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=131,
  serialized_end=823,
)


_MEASUREMENT = _descriptor.Descriptor(
  name='Measurement',
  full_name='openhtf.Measurement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='openhtf.Measurement.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='status', full_name='openhtf.Measurement.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='openhtf.Measurement.description', index=2,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='important', full_name='openhtf.Measurement.important', index=3,
      number=18, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='parameter_tag', full_name='openhtf.Measurement.parameter_tag', index=4,
      number=16, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numeric_value', full_name='openhtf.Measurement.numeric_value', index=5,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numeric_minimum', full_name='openhtf.Measurement.numeric_minimum', index=6,
      number=12, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numeric_maximum', full_name='openhtf.Measurement.numeric_maximum', index=7,
      number=13, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='text_value', full_name='openhtf.Measurement.text_value', index=8,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='expected_text', full_name='openhtf.Measurement.expected_text', index=9,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_optional', full_name='openhtf.Measurement.is_optional', index=10,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='set_time_millis', full_name='openhtf.Measurement.set_time_millis', index=11,
      number=19, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unit_code', full_name='openhtf.Measurement.unit_code', index=12,
      number=20, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_unit_code', full_name='openhtf.Measurement.custom_unit_code', index=13,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_unit_suffix', full_name='openhtf.Measurement.custom_unit_suffix', index=14,
      number=23, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='associated_attachment', full_name='openhtf.Measurement.associated_attachment', index=15,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(5000, 5200), ],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=1248,
)


_EVENTATTACHMENT = _descriptor.Descriptor(
  name='EventAttachment',
  full_name='openhtf.EventAttachment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='openhtf.EventAttachment.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value_binary', full_name='openhtf.EventAttachment.value_binary', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='openhtf.EventAttachment.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='set_time_millis', full_name='openhtf.EventAttachment.set_time_millis', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='parameter_tag', full_name='openhtf.EventAttachment.parameter_tag', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='openhtf.EventAttachment.type', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='openhtf.EventAttachment.mime_type', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='datatype', full_name='openhtf.EventAttachment.datatype',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1251,
  serialized_end=1447,
)

_MFGEVENT_BYLOT.containing_type = _MFGEVENT
_MFGEVENT.fields_by_name['lot'].message_type = _MFGEVENT_BYLOT
_MFGEVENT.fields_by_name['test_status'].enum_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._STATUS
_MFGEVENT.fields_by_name['assembly_events'].message_type = openhtf_dot_output_dot_proto_dot_assembly__event__pb2._ASSEMBLYEVENT
_MFGEVENT.fields_by_name['timings'].message_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._TIMING
_MFGEVENT.fields_by_name['phases'].message_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._PHASE
_MFGEVENT.fields_by_name['measurement'].message_type = _MEASUREMENT
_MFGEVENT.fields_by_name['attachment'].message_type = _EVENTATTACHMENT
_MFGEVENT.fields_by_name['test_logs'].message_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._TESTRUNLOGMESSAGE
_MFGEVENT.fields_by_name['failure_codes'].message_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._FAILURECODE
_MFGEVENT.oneofs_by_name['dut'].fields.append(
  _MFGEVENT.fields_by_name['dut_serial'])
_MFGEVENT.fields_by_name['dut_serial'].containing_oneof = _MFGEVENT.oneofs_by_name['dut']
_MFGEVENT.oneofs_by_name['dut'].fields.append(
  _MFGEVENT.fields_by_name['lot'])
_MFGEVENT.fields_by_name['lot'].containing_oneof = _MFGEVENT.oneofs_by_name['dut']
_MEASUREMENT.fields_by_name['status'].enum_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._STATUS
_MEASUREMENT.fields_by_name['unit_code'].enum_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._UNITS_UNITCODE
_EVENTATTACHMENT.fields_by_name['type'].enum_type = openhtf_dot_output_dot_proto_dot_test__runs__pb2._INFORMATIONTAG
_EVENTATTACHMENT.oneofs_by_name['datatype'].fields.append(
  _EVENTATTACHMENT.fields_by_name['type'])
_EVENTATTACHMENT.fields_by_name['type'].containing_oneof = _EVENTATTACHMENT.oneofs_by_name['datatype']
_EVENTATTACHMENT.oneofs_by_name['datatype'].fields.append(
  _EVENTATTACHMENT.fields_by_name['mime_type'])
_EVENTATTACHMENT.fields_by_name['mime_type'].containing_oneof = _EVENTATTACHMENT.oneofs_by_name['datatype']
DESCRIPTOR.message_types_by_name['MfgEvent'] = _MFGEVENT
DESCRIPTOR.message_types_by_name['Measurement'] = _MEASUREMENT
DESCRIPTOR.message_types_by_name['EventAttachment'] = _EVENTATTACHMENT

MfgEvent = _reflection.GeneratedProtocolMessageType('MfgEvent', (_message.Message,), dict(

  ByLot = _reflection.GeneratedProtocolMessageType('ByLot', (_message.Message,), dict(
    DESCRIPTOR = _MFGEVENT_BYLOT,
    __module__ = 'openhtf.output.proto.mfg_event_pb2'
    # @@protoc_insertion_point(class_scope:openhtf.MfgEvent.ByLot)
    ))
  ,
  DESCRIPTOR = _MFGEVENT,
  __module__ = 'openhtf.output.proto.mfg_event_pb2'
  # @@protoc_insertion_point(class_scope:openhtf.MfgEvent)
  ))
_sym_db.RegisterMessage(MfgEvent)
_sym_db.RegisterMessage(MfgEvent.ByLot)

Measurement = _reflection.GeneratedProtocolMessageType('Measurement', (_message.Message,), dict(
  DESCRIPTOR = _MEASUREMENT,
  __module__ = 'openhtf.output.proto.mfg_event_pb2'
  # @@protoc_insertion_point(class_scope:openhtf.Measurement)
  ))
_sym_db.RegisterMessage(Measurement)

EventAttachment = _reflection.GeneratedProtocolMessageType('EventAttachment', (_message.Message,), dict(
  DESCRIPTOR = _EVENTATTACHMENT,
  __module__ = 'openhtf.output.proto.mfg_event_pb2'
  # @@protoc_insertion_point(class_scope:openhtf.EventAttachment)
  ))
_sym_db.RegisterMessage(EventAttachment)


# @@protoc_insertion_point(module_scope)
